import json
import platform
import numpy as np
import gc

from time import time

from naeural_core import constants as ct

from collections import deque, defaultdict
from naeural_core import DecentrAIObject
from naeural_core.utils.sys_mon import SystemMonitor

MAX_LOG_SIZE = 10_000

MAX_LOCAL_HISTORY = int(30 * 60 / 10) # 30 minutes

BASE_INCREMENT_MB = 250


MIN_GLOBAL_INCREASE_GB = BASE_INCREMENT_MB * 2 / 1000
MIN_PROCESS_INCREASE_MB = BASE_INCREMENT_MB
MIN_PROCESS_INCREASE_GB = MIN_PROCESS_INCREASE_MB / 1000

WARMUP_READINGS = 5

class ApplicationMonitor(DecentrAIObject):
  def __init__(self, log, owner, **kwargs):
    self.owner = owner
    self.avail_memory_log = deque(maxlen=MAX_LOG_SIZE)
    self.process_memory_log = deque(maxlen=MAX_LOG_SIZE)
    self._process_memory_log_inited = False
    self._avail_memory_log_inited = False
    self.gpu_log = {}
    self.cpu_log = deque(maxlen=MAX_LOG_SIZE)
    self.start_avail_memory = None
    self._max_avail_increase = MIN_GLOBAL_INCREASE_GB
    self._max_process_increase = MIN_PROCESS_INCREASE_MB
    self.start_process_memory = None
    self.alert = False
    self.alert_avail = 0 # 0: no alert, 1: increase alert, 2: overall low mem alert    
    self.alert_process = False
    self.critical_alert = False
    self.owner_mem = -1
    self.__first_payload_prepared = False
    self.__first_ram_alert_raised = False
    self.__first_ram_alert_raised_time = 0
    self.__logged_hb_with_pipelines_status = False
    self.__logged_hb_with_active_plugins_status = False
    self.__last_delivered_error_notifs = defaultdict(int)
    self._done_first_smi_error = False
    self.dct_curr_nr = defaultdict(lambda:0)
    self.__last_temperature_info = None
    self.__local_data_history = {
      'cpu_load'            : deque(maxlen=MAX_LOCAL_HISTORY),
      'cpu_temp'            : deque(maxlen=MAX_LOCAL_HISTORY),
      'total_memory'        : deque(maxlen=MAX_LOCAL_HISTORY),
      'occupied_memory'     : deque(maxlen=MAX_LOCAL_HISTORY),

      'gpu_load'            : deque(maxlen=MAX_LOCAL_HISTORY),
      'gpu_temp'            : deque(maxlen=MAX_LOCAL_HISTORY),
      'gpu_occupied_memory' : deque(maxlen=MAX_LOCAL_HISTORY),
      'gpu_total_memory'    : deque(maxlen=MAX_LOCAL_HISTORY),
      
      'timestamps'          : deque(maxlen=MAX_LOCAL_HISTORY),      
    }
    super(ApplicationMonitor, self).__init__(log=log, prefix_log='[AMON]', **kwargs)
    return
  
  @property
  def first_payload_prepared(self):
    return self.__first_payload_prepared
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.STATUS
    super().P(s, prefix=False, color=color, **kwargs)
    return      
  
  def startup(self):
    super().startup()
    self.sys_mon = SystemMonitor(log=self.log, monitored_prefix=ct.THREADS_PREFIX, name='S_SysMon')
    if self.owner.cfg_system_monitor:
      self.sys_mon.start()
    
    if self.owner.cfg_system_temperature_check:
      self.P("INFO: System temperatures strict checking enabled...", color='r')
      self.get_temperature_status()
    else:
      self.P("WARNING: System temperatures checking disabled...", color='r')    
    return
  
  
  def get_temperature_status(self, return_max=True):
    temperature_info = self.log.get_temperatures(as_dict=True)
    self.__last_temperature_info = temperature_info
    temps = temperature_info['temperatures']
    msg = temperature_info['message']
    max_temp = None
    critical_temp = False
    max_val = 0
    max_sensor = ''
    if temps in [None, {}]:
      if self.owner.cfg_system_temperature_check:
        self.P(msg, color='r')
      temps = None
    else:
      for sensor, status in temps.items():
        current = status['current']
        critical_temp = status['current'] >= status['high']
        if max_val < current or critical_temp:
          max_val = current
          max_temp = "{} at {}°C".format(sensor, current)
          max_sensor = sensor
          if critical_temp:
            self.P("ALERT: High temperature detected: {}".format(max_temp), color='r')
            max_temp += ' ERROR/CRITICAL!'
            break
          #endif high temp
        #endfor values
    #endif temps
    self.__last_temperature_info['max_temp'] = max_val
    self.__last_temperature_info['max_temp_sensor'] = max_sensor
    if return_max:      
      return max_temp    
    else:        
      return temps
  
  
  def get_opencv_info(self):
    result = ""
    try:
      import cv2
      result = cv2.getBuildInformation()
    except:
      pass
    return result
  
  def is_ram_alert(self):
    is_alert = self.avail_memory_log[-1] < (self.owner.cfg_min_avail_mem_thr * self.log.total_memory)
    if is_alert and not self.__first_ram_alert_raised:
      self.__first_ram_alert_raised = True
      self.__first_ram_alert_raised_time = self.log.time_to_str()
      self.P("ALERT: First memory alert raised. Dumping logs", color='r')
      stats = gc.get_stats()
      self.P("GC stats:\n{}".format(json.dumps(stats, indent=4)))      
      self.P("GC garbage:\n{}".format(json.dumps(gc.garbage, indent=4)))
      self.owner._save_object_tree(
        fn='obj_mem_alert.txt', 
        save_top_only=True, 
        top=150
      )
      gc.collect()
    return is_alert
  
  
  def is_critical_ram_alert(self):
    is_critical_alert = False
    if len(self.avail_memory_log) >= 2:
      is_critical_alert = self.avail_memory_log[-1] < (self.owner.cfg_critical_restart_low_mem * self.log.total_memory)
    return is_critical_alert
      
  
  @property
  def slow_process_memory_log(self):
    result = []
    l = len(self.process_memory_log)
    if l > 0:
      for i in list(reversed(list(range(1,100,10)))):
        if l > i:
          result.append(round(self.process_memory_log[-i], 4))    
    return result
  
  def get_owner_version(self):
    app_ver = self.owner.__version__
    core_ver = self.owner.core_version

    _ver = '{}{}{}'.format(
      '{} | '.format(app_ver) if core_ver is not None else '',
      '{} | '.format(app_ver) if core_ver is None else '{} |'.format(core_ver),
      '{}'.format(self.log.version),
    )
    return _ver
  
  def log_status(self, color=None):
    if self.start_process_memory is None or self.start_avail_memory is None:      
      return
    delta_avail_start = self.start_avail_memory - self.avail_memory_log[-1]
    delta_avail_last = self.avail_memory_log[-2] - self.avail_memory_log[-1]
    delta_process_start = self.process_memory_log[-1] - self.start_process_memory
    delta_process_last =  self.process_memory_log[-1] -  self.process_memory_log[-2]
    avail_mem = self.avail_memory_log[-1]
    total_mem = self.log.total_memory
        
    np_a = np.array(self.process_memory_log)
    avg = np.mean(np_a[1:] - np_a[:-1])

    text = 'OK'
    if self.is_ram_alert():
      color = 'r'
      text = ' * * * LOW MEMORY: {:>6,.2f} < {:>6,.2f} ({} x {:>6,.2f}) Initial: {} * * * '.format(
        self.avail_memory_log[-1],  self.owner.cfg_min_avail_mem_thr * self.log.total_memory, self.owner.cfg_min_avail_mem_thr, self.log.total_memory,
        self.__first_ram_alert_raised_time,
      )
      # Check critical alert
      self.critical_alert = self.is_critical_ram_alert()
      # end critical alerting
    elif color is not None and color[0] == 'y':
      text = 'Potential memory leak'
 
      

    str_log  = "Memory status for '{}' <{}> (aa/ap: {}/{}):".format(      
      self.owner.cfg_eeid, self.get_owner_version(),
      self.alert_avail, self.alert_process,
    )
    str_log += "\n\r==========================================================================================="
    str_log += "\n\r==========================================================================================="
    str_log += "\n\r                       Status: {}".format(text + ' ({} readings)'.format(len(self.avail_memory_log)))
    str_log += "\n\r                       Global mem status/increase:"
    str_log += "\n\r                         Avail / Total             : {:>6,.2f} / {:>6,.2f} (GB)".format(avail_mem, total_mem)
    str_log += "\n\r                         Avail Start / Last decr.  : {:>6,.2f} / {:>6,.2f} (GB)".format(self.start_avail_memory, delta_avail_last)
    str_log += "\n\r                         Out-of-process mem loss   : {:>6,.2f} (GB)".format(delta_avail_start - delta_process_start)
    str_log += "\n\r                        -------------------------------------------------------------"
    str_log += "\n\r                       Process memory status and increase:"
    str_log += "\n\r                         Start / Now               : {:>6,.2f} / {:>6,.2f} (GB)".format(self.start_process_memory, self.process_memory_log[-1])
    str_log += "\n\r                         Inc / Last Inc / Mean Inc : {:>6,.2f} / {:>6,.2f} / {:6,.2f} (GB)".format(delta_process_start, delta_process_last, avg)
    if self.owner_mem > 0:
      str_log += "\n\r                         Object tree mem eval      :{:>3,.3f} GB".format(self.owner_mem / 1024**3)
    str_log += "\n\r==========================================================================================="
    str_log += "\n\r==========================================================================================="
    self.P(str_log, color=color)
    return
      
  
  def log_avail_memory(self):
    avail_memory = self.log.get_avail_memory(gb=True)
    self.avail_memory_log.append(avail_memory)
    if not self._avail_memory_log_inited:
      if len(self.avail_memory_log) >= WARMUP_READINGS:
        self.start_avail_memory = avail_memory
        self._avail_memory_log_inited = True
    else:
      if avail_memory < (self.owner.cfg_min_avail_mem_thr * self.log.total_memory):
        self.alert_avail = 2
      else:
        self.alert_avail = 0 
        if avail_memory < self.start_avail_memory:
          # check for abnormal increase
          delta_start = self.start_avail_memory - avail_memory        
          if delta_start >= (self._max_avail_increase + MIN_GLOBAL_INCREASE_GB):
            self.alert_avail = 1
            self._max_avail_increase = delta_start
          #endif
        #endif
      #endif
    #endif
    return avail_memory


  def log_process_memory(self):
    self.log.start_timer('log_proc_mem', section='app_mon')
    process_memory = self.log.get_current_process_memory(mb=False)
    self.process_memory_log.append(process_memory)
    if not self._process_memory_log_inited:
      if len(self.process_memory_log) >= WARMUP_READINGS :
        self.start_process_memory = process_memory
        self._process_memory_log_inited = True
    else:
      if process_memory > self.start_process_memory:
        # check for abnormal increase
        delta_start = process_memory - self.start_process_memory 
        if delta_start > (self._max_process_increase + MIN_GLOBAL_INCREASE_GB):
          self.alert_process = True
          self._max_process_increase = delta_start
        else:
          self.alert_process = False

    if False:
      # self.P("Calculating overall process memory...")
      self.log.start_timer('get_obj_size', section='app_mon')    
      self.owner_mem = self.log.get_obj_size(self.owner)
      self.log.end_timer('get_obj_size', section='app_mon')
      # self.P("Done calculating overall process memory.")
    self.log.end_timer('log_proc_mem', section='app_mon')
    return process_memory


  def add_data_info(self, val, stage):
    self.dct_curr_nr[stage] += val
    return
  
  def get_basic_perf_info(self):
    str_info = ''
    str_info += '\n\n'+ 40 * '=' + ' Basic load Info ' + 40 * '='
    cpu_loads = np.array(self.cpu_log).astype('float16')
    str_cpu_loads = ' '.join('{:>4.1f}'.format(x) for x in cpu_loads[-10:])
    str_info += "\n  Used CPU: '{}'".format(self.log.get_processor_platform())
    str_info += "\n    Avg. CPU load:      {:>4.1f}% ({} %load)".format(
      np.mean(self.cpu_log),
      str_cpu_loads,
    )
    str_info += "\n    Proc mem over time: {} (GB))".format(self.slow_process_memory_log)
    cuda = self.owner.serving_manager.default_cuda if self.owner.serving_manager is not None else 'cuda:0'
    if isinstance(cuda, str):
      cuda = cuda.lower().replace('cuda:','')
      try:
        cuda = int(cuda)
      except:
        cuda = 0
    if cuda in self.gpu_log:
      dct_gpu = self.gpu_log[cuda]
      gpu_name = dct_gpu.get(ct.GPU_INFO.NAME)
      gpu_used = dct_gpu.get(ct.GPU_INFO.GPU_USED)
      gpu_mem = dct_gpu.get(ct.GPU_INFO.ALLOCATED_MEM)
      gpu_total_mem = dct_gpu.get(ct.GPU_INFO.TOTAL_MEM)
      gpu_temp = dct_gpu.get(ct.GPU_INFO.GPU_TEMP)
      fan_speed = dct_gpu.get(ct.GPU_INFO.GPU_FAN_SPEED)
      if (gpu_used is not None) and (None not in gpu_used):
        str_info += "\n  Used GPU: [{}] '{}'".format(cuda, gpu_name)
        cuda_loads = np.array(gpu_used).astype('float16')
        mem_loads = np.array(gpu_mem).astype('float16')
        gpu_temp = np.array(gpu_temp).astype('float16')
        str_cuda_loads = ' '.join('{:>4.1f}'.format(x) for x in cuda_loads[-10:])
        str_mem_loads = ' '.join('{:>4.1f}'.format(x) for x in mem_loads[-10:])
        str_gpu_temp = ' '.join('{:>4.1f}'.format(x) for x in gpu_temp[-10:])
        str_info += "\n    Avg. GPU core used: {:>4.1f}% ({} %load)".format(
          np.mean(gpu_used),
          str_cuda_loads,
        )
        str_info += "\n    Avg. GPU mem load:  {:>4.1f}% ({} GB)".format(
          np.mean(gpu_mem) / gpu_total_mem * 100,
          str_mem_loads,
        )
        str_info += "\n    GPU temp:           {:>4.1f}°C ({} °C)".format(
          np.mean(gpu_temp), 
          str_gpu_temp
        )
      else:
        str_info += "\n  GPU issue: No GPU present or driver issue"
    return str_info

  def get_summary_perf_info(self, last_n=10):
    gpu_info_list = []

    cpu_mean = round(np.mean(list(self.cpu_log)[-last_n:]), 1)
    ee_occupied_mem = round(self.process_memory_log[-1], 1)
    pc_occupied_mem = round(self.log.total_memory - self.avail_memory_log[-1], 1)
    pc_total_mem = round(self.log.total_memory, 1)

    gpu_load, gpu_total_mem, gpu_occupied_memory, gpu_temp, gpu_fan, gpu_fan_unit = None, None, None, None, None, None

    for i, dct_gpu in self.gpu_log.items():
      if gpu_load is None:
        lst_gpu_load = list(dct_gpu[ct.GPU_INFO.GPU_USED])
        if any(isinstance(x, str) or x is None for x in lst_gpu_load):
          lst_gpu_load = [-1] * len(lst_gpu_load)
        gpu_load = round(lst_gpu_load[-1], 1)

        gpu_total_mem = dct_gpu[ct.GPU_INFO.TOTAL_MEM]
        gpu_total_mem = round(gpu_total_mem, 1) if isinstance(gpu_total_mem, (int, float)) else None

        lst_gpu_occupied_memory = list(dct_gpu[ct.GPU_INFO.ALLOCATED_MEM])
        if any(isinstance(x, str) or x is None for x in lst_gpu_occupied_memory):
          lst_gpu_occupied_memory = [-1] * len(lst_gpu_occupied_memory)
        gpu_occupied_memory = round(lst_gpu_occupied_memory[-1], 1)

        lst_gpu_temp = list(dct_gpu[ct.GPU_INFO.GPU_TEMP])
        if any(isinstance(x, str) or x is None for x in lst_gpu_temp):
          lst_gpu_temp = [-1] * len(lst_gpu_temp)
        gpu_temp = round(lst_gpu_temp[-1], 1)

        lst_gpu_fan = list(dct_gpu[ct.GPU_INFO.GPU_FAN_SPEED])
        if any(isinstance(x, str) or x is None for x in lst_gpu_fan):
          lst_gpu_fan = [-1] * len(lst_gpu_fan)
        gpu_fan = round(lst_gpu_fan[-1], 1)

        gpu_fan_unit = dct_gpu[ct.GPU_INFO.GPU_FAN_SPEED_UNIT]

      gpu_load_mean = round(np.mean(lst_gpu_load[-last_n:]), 1)
      str_gpu_info = "cuda:{} {}%, mem {}/{}GB, {}°C, gpu fan {}{}".format(
        i,
        gpu_load_mean,
        gpu_occupied_memory,
        gpu_total_mem,
        gpu_temp,
        gpu_fan,
        gpu_fan_unit
      )
      gpu_info_list.append(str_gpu_info)

    str_gpu_info_list = ", ".join(gpu_info_list) if len(gpu_info_list) > 0 else "no gpu"

    str_info = "cpu {}%, ram(EE) {}({})/{} GB, {}".format(
      cpu_mean,
      pc_occupied_mem, ee_occupied_mem,
      pc_total_mem,
      str_gpu_info_list
    )
    
    temp = self.get_temperature_status(return_max=True)
    cpu_temp = None
    if temp is not None:
      str_info += ", {}".format(temp)
      cpu_temp = self.__last_temperature_info['max_temp']

    self.add_local_history_and_save(
      cpu_load=cpu_mean,
      cpu_temp=cpu_temp,
      total_memory=pc_total_mem,
      occupied_memory=pc_occupied_mem,
      gpu_load=gpu_load,
      gpu_temp=gpu_temp,
      gpu_fan=gpu_fan,
      gpu_fan_unit=gpu_fan_unit,
      gpu_occupied_memory=gpu_occupied_memory,
      gpu_total_memory=gpu_total_mem,
      timestamps=self.log.now_str(nice_print=True, short=False),
    )

    return str_info

  def _add_gpu_data(self, lst_info):
    if not isinstance(lst_info, list):
      return
    for i,gpu in enumerate(lst_info):
      if i not in self.gpu_log:
        self.gpu_log[i] = {
          ct.GPU_INFO.TOTAL_MEM : gpu[ct.GPU_INFO.TOTAL_MEM],
          ct.GPU_INFO.NAME : gpu[ct.GPU_INFO.NAME],
          ct.GPU_INFO.GPU_USED : deque(maxlen=MAX_LOG_SIZE),
          ct.GPU_INFO.ALLOCATED_MEM : deque(maxlen=MAX_LOG_SIZE),
          ct.GPU_INFO.GPU_TEMP : deque(maxlen=MAX_LOG_SIZE),
          ct.GPU_INFO.GPU_FAN_SPEED : deque(maxlen=MAX_LOG_SIZE),
          ct.GPU_INFO.GPU_FAN_SPEED_UNIT : gpu[ct.GPU_INFO.GPU_FAN_SPEED_UNIT],
        }
      self.gpu_log[i][ct.GPU_INFO.GPU_USED].append(gpu[ct.GPU_INFO.GPU_USED])
      self.gpu_log[i][ct.GPU_INFO.ALLOCATED_MEM].append(gpu[ct.GPU_INFO.ALLOCATED_MEM])
      self.gpu_log[i][ct.GPU_INFO.GPU_TEMP].append(gpu[ct.GPU_INFO.GPU_TEMP])
      self.gpu_log[i][ct.GPU_INFO.GPU_FAN_SPEED].append(gpu[ct.GPU_INFO.GPU_FAN_SPEED])      
    return
      

  def get_gpu_info(self):
    # self.log.start_timer('gpu_info', section='async_comm')
    info = self.log.gpu_info(mb=False) or "Pytorch or NVidia-SMI issue"
    # self.log.end_timer('gpu_info', skip_first_timing=False, section='async_comm')
    
    if isinstance(info, list):
      self._add_gpu_data(info)
    return info

  def maybe_send_error_notification(
    self, error_type, msg, info, displayed=True, time_threshold=300
  ):
    last_send = self.__last_delivered_error_notifs[error_type]
    if (time() - last_send) > time_threshold:
      self.__last_delivered_error_notifs[error_type] = time()
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        msg=msg, 
        info=info,
        displayed=displayed,
      )
    return
  
  def add_local_history_and_save(self, **kwargs):
    # TODO: refactor this code to use constants
    dct_to_save = {
      'address' : self.owner.e2_addr,
      'eth_address' : self.owner.eth_address,
      'alias' : self.owner.cfg_eeid,
    }    
    current_epoch = self.owner._network_monitor.epoch_manager.get_current_epoch()
    dct_to_save['current_epoch'] = current_epoch
    dct_to_save['current_epoch_avail'] = self.owner._network_monitor.epoch_manager.get_current_epoch_availability()
    dct_to_save['uptime'] = self.log.elapsed_to_str(self.owner.running_time)
    dct_to_save['version'] = self.get_owner_version().split(' ')[0]
    epochs = self.owner._network_monitor.epoch_manager.get_node_epochs(
      node_addr=self.owner.e2_addr, autocomplete=True, as_list=True
    )
    if isinstance(epochs, list):
      dct_to_save['last_epochs'] = epochs[-10:]
    else:
      dct_to_save['last_epochs'] = []
    #endif epochs
    dct_to_save['last_save_time'] = self.log.now_str(nice_print=True, short=True)

    for k,v in kwargs.items():
      if k in self.__local_data_history:
        self.__local_data_history[k].append(v)
        dct_to_save[k] = list(self.__local_data_history[k])

    self.log.save_json_to_data(dct_to_save, 'local_history.json', indent=False)    
    return
  

  def get_status(self, status=None, full=True, send_log=False):
    machine_ip = self.log.get_localhost_ip()
    machine_memory = round(self.log.total_memory,3)
    cpu_usage = self.log.get_cpu_usage()
    self.cpu_log.append(cpu_usage)
    
    is_docker = self.owner.runs_in_docker
    docker_env = self.owner.docker_env
    docker_source = self.owner.docker_source
    
    # check & get mem statuses required for log_status
    avail_memory = round(self.log_avail_memory(),3)
    process_memory = round(self.log_process_memory(),3)
    # log status if required
    if (not self.alert_avail) and (not self.alert_process):
      self.alert = False
    else:
      self.log_status(color=None if self.alert_avail == 0 else 'r')    
    
    
    # hard check avail disk
    # TODO: automatic checker of info data post gathering
    total_disk = round(self.log.total_disk, 3)
    avail_disk = round(self.log.get_avail_disk(gb=True), 3)
    display = True
    if avail_disk < self.owner.cfg_min_avail_disk_thr:
      info = "WARNING: Avail disk on '{}' current volume is {:.1f} GB below minimum of {:.1f} GB".format(
        self.owner.cfg_eeid,
        avail_disk,
        self.owner.cfg_min_avail_disk_thr
      )
      if display:
        self.P(info, color='error')
        
      self.maybe_send_error_notification(
        error_type="disk",
        msg="LOW DISK SPACE ON '{}'".format(self.owner.cfg_eeid),
        info=info,
        displayed=display,          
      )
    
    if avail_memory < self.owner.cfg_min_avail_mem_thr * self.log.total_memory:
      info = "WARNING: available memory {:.2f} GB ({:.1f}%) below {:.2f} GB ({:.1f}%) threshold of total memory ({:.1f} GB)".format(
        avail_memory, avail_memory / self.log.total_memory * 100,
        self.owner.cfg_min_avail_mem_thr * self.log.total_memory, self.owner.cfg_min_avail_mem_thr * 100,
        self.log.total_memory,
      )
      if display:
        self.P(info, color='error')

      self.maybe_send_error_notification(
        error_type="memory",
        msg="LOW MEMORY ON '{}'".format(self.owner.cfg_eeid),
        info=info,
        displayed=display,
      )
      
    # now get basic info
    
    str_cpu = self.log.get_processor_platform()
    nr_cores = self.log.get_processor_cores()
    
    total_time = self.owner.running_time
    timestamp = self.log.now_str(nice_print=False)
    current_time = self.log.now_str(nice_print=True, short=False) # leave with milis - modify downstream if necessary

    n_payloads = self.dct_curr_nr[ct.NR_PAYLOADS]
    n_inferences = self.dct_curr_nr[ct.NR_INFERENCES]
    n_stream_data = self.dct_curr_nr[ct.NR_STREAMS_DATA]
    self.dct_curr_nr = defaultdict(lambda:0)
    
    if status is None:
      status = ct.DEVICE_STATUS_ONLINE
    
    lst_gpus = self.get_gpu_info()
    
    loops_timings = {}
    if self.owner.comm_manager is not None:
      loops_timings = {
        'main_loop_avg_time' : self.owner.avg_loop_timings,
        'comm_loop_avg_time' : self.owner.comm_manager.avg_comm_loop_timings,
      }

    active_serving_processes = []
    serving_pids = []
    default_cuda = ''
    if self.owner.serving_manager is not None:
      active_serving_processes = self.owner.serving_manager.get_active_servers(show=False)
      default_cuda = self.owner.serving_manager.default_cuda
      serving_pids = self.owner.serving_manager.serving_pids
    str_active_serving_processes = '\n\nActive serving processes (in-process: {}):'.format(self.owner.in_process_serving)  
    if len(active_serving_processes) == 0:
      str_active_serving_processes += '\n  --  No running serving processes --'
    for svr in active_serving_processes:
      str_active_serving_processes += '\n'
      for k,v in svr.items():
        str_active_serving_processes += "\n    {:<15}{}".format(k + ':', v)

    str_gpu_info = self.get_basic_perf_info()
    
    str_timers = None
    if full:
      str_timers = '\n'.join(self.log.format_timers())
      # now get the sysmon info
      str_timers += '\n\n' + '\n'.join(self.get_sys_mon_info())
      # add package info
      str_timers +='\n\nPackage information (Python {}):\n{}'.format(
        self.log.python_version,
        self.log.get_packages(as_text=True, indent=4)
      )
      # add serving processes
      str_timers += str_active_serving_processes
      # add open-cv if available
      str_timers += "\n\n" + self.get_opencv_info()
    #endif prepare timers if necessary
    
    dct_cap_info = None
    if self.owner.capture_manager is not None:
      dct_cap_info = self.owner.capture_manager.get_captures_status(display=False, as_dict=True)
    
    dct_comm_info = None  
    inkB, outkB = 0, 0
    if self.owner.comm_manager is not None:
      dct_comm_info = self.owner.comm_manager.get_comms_status()
      for _, dct_c in dct_comm_info.items():
        inkB += dct_c[ct.HB.COMM_INFO.IN_KB]
        outkB += dct_c[ct.HB.COMM_INFO.OUT_KB]
      
    str_git_branch = self.sys_mon.git_branch
    str_conda_branch = self.log.conda_env
    
    if is_docker:
      str_git_branch = 'Docker:' + docker_source
      str_conda_branch = 'Env:' + docker_env      
    
    if send_log:
      error_log = self.log.err_log
      dev_log = self.log.app_log
    else:
      dev_log   = 'Device log available only upon request or in special situations'
      error_log = 'Error log available only upon request or in special situations'
    #endif send log

    hb_contains_pipelines = self.owner.cfg_hb_contains_pipelines
    hb_contains_active_plugins = self.owner.cfg_hb_contains_active_plugins
    
    # The following is required to update the internal state of the business manager
    _active_plugin = self.owner.business_manager.get_active_plugins_instances(
        as_dict=True # send business plugin instances info as objects (NOT as lists)
      ) # this call is REQUIRED for internal state update
    active_business_plugins = []
    if self.owner.business_manager is not None and hb_contains_active_plugins:
      active_business_plugins = _active_plugin
    

    s_os = platform.platform()
    s_os = s_os.replace('Windows', 'W').replace('Linux','L')
    _ver = '{} {}'.format(self.get_owner_version(), s_os)
        
    lst_config_streams = []    
    if self.owner.config_manager is not None and hb_contains_pipelines:      
      lst_config_streams = self.owner.get_pipelines_view()
      
    if (not self.__logged_hb_with_pipelines_status) and (self.owner.config_manager is not None):
      if hb_contains_pipelines:
        self.P(f"INFO: Heartbeat with pipelines status. Current: {len(lst_config_streams)}", boxed=True)
      else:
        self.P("INFO: Heartbeat WITHOUT pipelines status", boxed=True)
      self.__logged_hb_with_pipelines_status = True
      
      
    if (not self.__logged_hb_with_active_plugins_status) and (self.owner.business_manager is not None):
      if hb_contains_active_plugins:
        self.P(f"INFO: Heartbeat with active plugins status. Current: {len(active_business_plugins)}", boxed=True)
      else:
        self.P("INFO: Heartbeat WITHOUT active plugins status", boxed=True)
      self.__logged_hb_with_active_plugins_status = True
    
    

    address = self.owner.e2_addr      
    is_supervisor = self.owner.is_supervisor_node
    
    whitelist = self.owner.whitelist
    
    

    dct_status = {
      #mandatory
      ct.HB.CURRENT_TIME      : current_time,
      
      ct.HB.EE_ADDR           : address,
      ct.HB.EE_WHITELIST      : whitelist,
      ct.HB.EE_IS_SUPER       : is_supervisor,
      ct.HB.EE_FORMATTER      : self.owner.cfg_io_formatter,
      
      # ALERTS
      ct.HB.IS_ALERT_RAM      : self.is_ram_alert(),
      # End ALERTS

      ct.HB.NR_INFERENCES     : n_inferences,
      ct.HB.NR_PAYLOADS       : n_payloads,
      ct.HB.NR_STREAMS_DATA   : n_stream_data,
      ct.HB.PY_VER            : self.log.python_version,

      ct.HB.SECURED           : self.owner.is_secured,

      ct.HB.EE_HB_TIME        : self.owner.cfg_app_seconds_heartbeat,
      ct.HB.DEVICE_STATUS     : status,
      ct.HB.MACHINE_IP        : machine_ip,
      ct.HB.MACHINE_MEMORY    : machine_memory,
      ct.HB.AVAILABLE_MEMORY  : avail_memory,
      ct.HB.PROCESS_MEMORY    : process_memory,      
      ct.HB.CPU_USED          : cpu_usage,
      ct.HB.CPU_NR_CORES      : nr_cores,
      ct.HB.GPUS              : lst_gpus,
      ct.HB.GPU_INFO          : str_gpu_info,
      ct.HB.DEFAULT_CUDA      : default_cuda,
      ct.HB.CPU               : str_cpu,
      ct.HB.TIMESTAMP         : timestamp,
      ct.HB.UPTIME            : total_time,
      ct.HB.VERSION           : _ver,
      ct.HB.LOGGER_VERSION    : self.log.version,
      ct.HB.TOTAL_DISK        : total_disk,
      ct.HB.AVAILABLE_DISK    : avail_disk,
      ct.HB.ACTIVE_PLUGINS    : active_business_plugins,
      ct.HB.STOP_LOG          : self.owner.main_loop_stop_log,
      
      ct.HB.R1FS_ID           : self.owner.r1fs_id,
      ct.HB.R1FS_ONLINE       : self.owner.r1fs_started,
      ct.HB.R1FS_RELAY        : self.owner.r1fs_relay,  
      
      ct.HB.TEMPERATURE_INFO  : self.__last_temperature_info,
      
      ct.HB.GIT_BRANCH        : str_git_branch,
      ct.HB.CONDA_ENV         : str_conda_branch,      
        
      ct.HB.CONFIG_STREAMS    : lst_config_streams,
      ct.HB.DCT_STATS         : dct_cap_info,
      ct.HB.COMM_STATS        : dct_comm_info,
      ct.HB.COMM_INFO.IN_KB   : round(inkB, 3),
      ct.HB.COMM_INFO.OUT_KB  : round(outkB, 3),
        
      ct.HB.SERVING_PIDS      : serving_pids, 
      ct.HB.LOOPS_TIMINGS     : loops_timings,

      ct.HB.TIMERS            : 'Timers not available in summary',
      ct.HB.DEVICE_LOG        : dev_log,
      ct.HB.ERROR_LOG         : error_log
    }
    
    dct_ext_status = {    
      ct.HB.TIMERS            : str_timers,
    }
    
    
    if full:
      for k, v in dct_ext_status.items():
        dct_status[k] = v
        
    if self.owner.cfg_compress_heartbeat:
      str_text = "{}"
      try:
        str_text = json.dumps(dct_status)
      except Exception as exc:
        self.P("ERROR: Could not compress heartbeat: {}\n{}".format(exc, dct_status), color='error')
      str_encoded = self.log.compress_text(str_text)
      dct_status = {
        ct.HB.ENCODED_DATA : str_encoded,
        # needed in caller...
        ct.HB.NR_INFERENCES     : n_inferences,
        ct.HB.NR_PAYLOADS       : n_payloads,
        ct.HB.NR_STREAMS_DATA   : n_stream_data,        
        ct.HB.HEARTBEAT_VERSION : ct.HB.V2, 
      }
    else:
      dct_status[ct.HB.HEARTBEAT_VERSION] = ct.HB.V1
      
    if not self.first_payload_prepared:
      self.__first_payload_prepared = True
      self.P("First hb for {}".format(address), boxed=True)

    return dct_status
  
  
  def get_sys_mon_info(self):
    if hasattr(self, 'sys_mon') and self.sys_mon is not None:
      return self.sys_mon.get_status_log(descending=True)
    return []

  def log_sys_mon_info(self, color='n'):
    info = self.get_sys_mon_info()
    s = "\n" + "\n".join(info)
    self.log.P(s, color=color)
    return


  def _maybe_stop_sys_mon(self):
    if hasattr(self, 'sys_mon') and self.sys_mon is not None:
      self.sys_mon.stop()
      self.sys_mon.log_status()
    return
  

  def shutdown(self):
    super().shutdown()
    self._maybe_stop_sys_mon()
    


from naeural_core.business.base import BasePluginExecutor as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,

  'ALLOW_EMPTY_INPUTS': True,

  'DATASET_OBJECT_NAME': None,
  'DATASET_LOCAL_PATH': None,
  'DELETE_AFTER_ZIP': False,
  'INCLUDE_DIR_IN_ZIP': False,

  'IS_RAW': False,

  'PLUGIN_LOOP_RESOLUTION': 1/10,


  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },

}


class MinioUploadDatasetPlugin(BasePlugin):
  CONFIG = _CONFIG

  def on_init(self):
    super(MinioUploadDatasetPlugin, self).on_init()
    self.__uploaded = False
    return

  def startup(self):
    super().startup()
    assert self.cfg_dataset_object_name is not None
    assert self.cfg_dataset_local_path is not None
    return

  def _process(self):
    if not self.__uploaded:
      local_path = self.cfg_dataset_local_path
      # In case the local path is a directory, zip it if it is not already zipped
      if self.os_path.isdir(local_path) and not self.os_path.exists(f'{local_path}.zip'):
        zip_path = self.diskapi_zip_dir(local_path, include_dir=self.cfg_include_dir_in_zip)
        if self.cfg_delete_after_zip:
          self.diskapi_delete_directory(local_path)
        local_path = zip_path
      # endif local_path is a directory and not zipped
      # In case the local path is a directory and the zip file exists, use the zip file
      local_path = local_path if local_path.endswith('.zip') else f'{local_path}.zip'
      target_path = self.cfg_dataset_object_name
      target_path = target_path if target_path.endswith('.zip') else f'{target_path}.zip'
      if self.os_path.exists(local_path):
        self.upload_file(
          file_path=local_path,
          target_path=target_path,
          force_upload=True,
        )
        self.__uploaded = True
        self.cmdapi_stop_current_stream()
      # endif local path exists
    # endif dataset uploaded
    # Report upload status if final dataset
    if not self.cfg_is_raw and self.__uploaded:
      self.add_payload_by_fields(
        is_final_dataset_status=not self.cfg_is_raw,
        uploaded=self.__uploaded
      )
    # endif report

    return



import urllib3
import minio
from minio import Minio


from naeural_core import constants as ct
from naeural_core.business.base import BasePluginExecutor

__VER__ = '0.2.2'

class MINIO_STATES:
  IN_PROC = 'IN_PROC'
  DONE = 'DONE'

_CONFIG = {
  **BasePluginExecutor.CONFIG,
  'MINIO_HOST'          : None,
  'MINIO_ACCESS_KEY'    : None,
  'MINIO_SECRET_KEY'    : None,
  'MINIO_SECURE'        : None,
  
  'MINIO_TIMEOUT'       : 20,
  
  'ENV_HOST'            : 'EE_MINIO_ENDPOINT',
  'ENV_ACCESS_KEY'      : 'EE_MINIO_ACCESS_KEY',
  'ENV_SECRET_KEY'      : 'EE_MINIO_SECRET_KEY',
  'ENV_SECURE'          : 'EE_MINIO_SECURE',

  'ALLOW_EMPTY_INPUTS'  : True,
  
  'PROCESS_DELAY'       : 30,
  
  'MINIO_IDLE_SECONDS'  : 7200, # total idle time after full inspection of all buckets
  
  'MAX_FILES_PER_ITER'  : 500,

  'ALERT_DATA_COUNT'    : 1,
  'ALERT_RAISE_VALUE'   : 0.8,
  'ALERT_LOWER_VALUE'   : 0.75,
  'ALERT_MODE'          : 'min',

  'MAX_SERVER_QUOTA'    : 90,
  
  'QUOTA_UNIT'          : ct.FILE_SIZE_UNIT.GB,
  
  'MINIO_DEBUG_MODE'    : True,
  
  'MIN_TIME_BETWEEN_PAYLOADS' : 60 * 5, # 5 minutes

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}

class MinioMonit01Plugin(BasePluginExecutor):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(MinioMonit01Plugin, self).__init__(**kwargs)
    return

  
  def on_init(self):
    infos = f"""
    - Minio v{minio.__version__}
    - Plugin version: {__VER__}
    - Configured server quota: {self.cfg_max_server_quota} {self.cfg_quota_unit}
    """
    if self.is_supervisor_node:
      self.P(f"{self.__class__.__name__} initializing on SUPERVISOR node: {infos}")
    else:
      self.P(f"{self.__class__.__name__} initializing on simple worker node: {infos}")
    self.__global_iter = 0
    self.__last_display = 0
    self.__errors_during_iteration = 0
    self.__minio_idle_wait_time = self.cfg_minio_idle_seconds
    
    self.__iter_start = self.time()
    
    
    self.__default_host = self.cfg_minio_host
    self.__default_access_key = self.cfg_minio_access_key
    self.__default_secret_key = self.cfg_minio_secret_key
    self.__default_secure = self.cfg_minio_secure

    self.__host = None
    self.__access_key = None
    self.__secret_key = None
    self.__secured = None


    self.__minio_client = None
    self.__plugin_state = None
    self.__idle_start = None
    self.__buckets_list = []
    self.__current_bucket_objects_generator = None
    self.__last_minio_payload_time = 0
    self.__reset_size()
    return

  
  def __reset_size(self):
    self.__server_size = 0
    self.__bucket_size = {}
    self.__current_bucket_no = 0
    self.__errors_during_iteration = 0
    return
  
  
  def show_minio_config(self, error=False):
    view_access = self.__access_key[:3] + "*" * (len(self.__access_key) - 3)
    view_secret = self.__secret_key[:3] + "*" * (len(self.__secret_key) - 3)
    infos = f"""
    MINIO VER:    {minio.__version__}
    MINIO_HOST:   {self.__default_host}
    SECURE:       {self.__default_secure}
    ACCESS_KEY:   {view_access}
    SECRET_KEY:   {view_secret}  
    SERVER_QUOTA: {self.cfg_max_server_quota} {self.cfg_quota_unit}
    VERBOSE:      {self.cfg_minio_debug_mode}
    """
    self.P(f"Minio config params: {infos}", color='r' if error else None)
    return
    
  
  
  def __maybe_get_env_config(self):
    if self.is_supervisor_node and self.__default_host is None:
      self.P("Configuring supervisor node using environment variables for Minio connection...")
      self.__default_host = self.os_environ.get(self.cfg_env_host)
      self.__default_access_key = self.os_environ.get(self.cfg_env_access_key)
      self.__default_secret_key = self.os_environ.get(self.cfg_env_secret_key)
      self.__default_secure = self.json_loads(self.os_environ.get(self.cfg_env_secure))
    #endif
    return
  
  
  def __get_connection_config(self):
    self.__maybe_get_env_config()
    self.__host = self.cfg_minio_host or self.__default_host
    self.__access_key = self.cfg_minio_access_key or self.__default_access_key
    self.__secret_key = self.cfg_minio_secret_key or self.__default_secret_key
    self.__secured = self.cfg_minio_secure or self.__default_secure
    return self.__host, self.__access_key, self.__secret_key, self.__secured
  
      

  def __get_next_bucket(self):
    if len(self.__buckets_list) == 0:
      # reset size and get new buckets
      self.__reset_size()
      self.__buckets_list = self.__minio_client.list_buckets()  
      for bucket in self.__buckets_list:
        self.__bucket_size[bucket.name] = {
          'size': 0,
          'objects': 0,
          'minio_errors' : 0,
        }        
      str_buckets = ", ".join(map(lambda x: x.name, self.__buckets_list))    
      DISPLAY_EVERY = 15 * 60 
      if (self.time() - self.__last_display) > DISPLAY_EVERY:
        self.__last_display = self.time()
        self.P("Analysing {} (secured: {}). Iterating through {} buckets: {}".format(
          self.__host, self.__secured,
          len(self.__buckets_list), 
          str_buckets,
          )
        )
    self.__current_bucket_no += 1
    return self.__buckets_list.pop()


  def __maybe_get_new_objects_generator(self):
    if self.__current_bucket_objects_generator is None:
      self.__current_bucket = self.__get_next_bucket()
      self.__current_bucket_objects_generator = self.__minio_client.list_objects(
        bucket_name=self.__current_bucket.name, 
        recursive=True,
      )
      if self.cfg_minio_debug_mode:
        self.P("Iteration {}/{} through bucket '{}'".format(
          self.__current_bucket_no, len(self.__bucket_size),
          self.__current_bucket.name, 
          )
        )
    return


  def __maybe_create_client_connection(self):
    host, access_key, secret_key, secured = self.__get_connection_config()    
    if (
      self.__minio_client is None and
      host is not None and
      access_key is not None and
      secret_key is not None and
      secured is not None
      ):
        self.P("Creating Minio client connection at iteration {} to {}...".format(self.__global_iter, host))
        self.show_minio_config()
        http_client = urllib3.PoolManager(
          timeout=urllib3.Timeout(
            connect=self.cfg_minio_timeout,             
            read=self.cfg_minio_timeout,
          ),
          cert_reqs='CERT_NONE'  # Disable SSL certificate verification
        )
        self.__minio_client = Minio(
          endpoint=host,
          access_key=access_key,
          secret_key=secret_key,
          secure=secured,
          http_client=http_client,
          cert_check=False,  # disable SSL certificate check   
        )
    else:
      if self.__minio_client is None and self.__global_iter < 2:
        self.P(f"Missing Minio connection parameters (is_supervisor_node: {self.is_supervisor_node}, host: {host}, access_key: {access_key}, secret_key: {secret_key}, secure: {secured})", color='r')
    return


  def __process_iter_files(self):
    MAX_ITER_TIME = 10
    start_time = self.time()
    max_nr_files = self.cfg_max_files_per_iter
    if self.cfg_minio_debug_mode:
      self.P("Processing batch of {} files in bucket '{}'".format(
        max_nr_files, self.__current_bucket.name
      ))
    local_count = 0
    for _ in range(max_nr_files):
      try:
        obj = next(self.__current_bucket_objects_generator)
        self.__server_size += obj.size
        local_count += 1
        self.__bucket_size[self.__current_bucket.name]['size'] += obj.size
        self.__bucket_size[self.__current_bucket.name]['objects'] += 1
        if (self.time() - start_time) > MAX_ITER_TIME:
          self.P("WARNING: iterated through {} files out of batch({}) in {:.1f}s on bucket {}".format(
            local_count, max_nr_files, self.time() - start_time, self.__current_bucket.name
          ), color='r')
          break
      except StopIteration:
        self.__current_bucket_objects_generator = None
        self.P("Finished processing bucket '{}'".format(self.__current_bucket.name))
        break
      except Exception as e:
        tb_info = self.trace_info()
        self.P("Error processing bucket '{}' at file no {} of batch({}): {}\n{}".format(
          self.__current_bucket.name, local_count, max_nr_files, e, tb_info), color='r'
        )
        self.show_minio_config(error=True)
        self.__errors_during_iteration += 1
        self.__bucket_size[self.__current_bucket.name]['minio_errors'] += 1
        break
    elapsed_time = self.time() - start_time + 1e-10
    if self.cfg_minio_debug_mode:
      if local_count > 0:
        self.P("  Processed {} objects in {:.1f}s, {:.0f} files/s (total time: {:.1f})".format(
          local_count, elapsed_time, local_count / elapsed_time, self.time() - self.__iter_start,) 
        )
        n_o = self.__bucket_size[self.__current_bucket.name]['objects']
        cut = self.cfg_max_files_per_iter * 3
        if n_o > 0 and (n_o % cut) == 0:
          self.P("  Bucket '{}' size: {:.2f} {} (so far for {} objs)".format(
            self.__current_bucket.name,
            self.convert_size(self.__bucket_size[self.__current_bucket.name]['size'], self.cfg_quota_unit),
            self.cfg_quota_unit,
            self.__bucket_size[self.__current_bucket.name]['objects'],
            )
          )
      else:
        self.P("  No objects processed in bucket '{}'".format(self.__current_bucket.name))
    return


  def _process(self):
    self.__global_iter += 1
    payload = None
    
    # if the plugin has finished a full cycle wait some time done, just return the payload
    if self.__plugin_state == MINIO_STATES.DONE:
      if (self.time() - self.__idle_start) > self.__minio_idle_wait_time:
        self.__plugin_state = MINIO_STATES.IN_PROC
        self.__idle_start = None
        self.__iter_start = self.time()
        self.P("Restarting S3 full inspection iteration..")
      else:
        return payload

    try:
      self.__maybe_create_client_connection()
      if self.__minio_client is None:
        return

      self.__maybe_get_new_objects_generator()
      self.__process_iter_files()
      
    except Exception as exc:
      SLEEP_TIME = 300
      self.show_minio_config(error=True)
      self.P(f"Error during MinIO Monitor: {exc}, blocking for {SLEEP_TIME}", color='r')
      self.sleep(SLEEP_TIME)
      return

    # when the plugin iterated through all buckets send payload
    if self.__current_bucket_objects_generator is None and len(self.__buckets_list) == 0:
      # set done state
      self.__plugin_state = MINIO_STATES.DONE
      self.__idle_start = self.time()
      
      converted_size = self.convert_size(self.__server_size, self.cfg_quota_unit)
      percentage_used = converted_size / self.cfg_max_server_quota

      self.alerter_add_observation(percentage_used)
      
      for b in self.__bucket_size:
        self.__bucket_size[b]['size_h'] = "{:,.2f} {}".format(
          self.convert_size(self.__bucket_size[b]['size'], self.cfg_quota_unit),
          self.cfg_quota_unit,
        )
      
      
      if (self.time() - self.__last_minio_payload_time) > self.cfg_min_time_between_payloads:
        self.__last_minio_payload_time = self.time()
        elapsed = self.time() - self.__iter_start
        msg = """Server size: 
        Total size: {:.1f} {} 
        Configured quota: {:.1f} {}
        Quota percentage: {:.1f} %
        Minio errors during iteration: {}
        Time during analysis: {:.1f}s
        Alerter: {}""".format(
          converted_size, self.cfg_quota_unit, self.cfg_max_server_quota, self.cfg_quota_unit,
          percentage_used * 100, self.__errors_during_iteration, elapsed,
          self.get_alerter_status(),
        )
        color = 'r' if self.alerter_is_alert() else None
        # only log-show detailed info in debug mode
        if self.cfg_minio_debug_mode:
          self.P("{}\nResults:\n{}".format(
            msg, self.json_dumps(self.__bucket_size, indent=2),
            ), color=color
          )
        else:
          self.P(msg, color=color)
        # alerts are handled automatically by above code: adding used percentage
        payload = self._create_payload(
          server_size=self.__server_size,
          buckets=self.__bucket_size,
          status=msg,
          minio_errors_during_analisys=self.__errors_during_iteration,
        )

      if self.__errors_during_iteration > 0:
        self.__minio_idle_wait_time = self.cfg_minio_idle_seconds // 2
        self.P("Errors during iteration: {}, lowering idle time to {}s".format(
          self.__errors_during_iteration, self.__minio_idle_wait_time), color='r'
        )
      else:
        self.__minio_idle_wait_time = self.cfg_minio_idle_seconds
      #endif time check
    return payload


"""
[Y5F][2021-10-08 08:26:09] Full process GPU memory amprent: 2.62 GB
[Y5F][2021-10-08 08:26:09] Timing results GPU_SYNC=True, TO_CPU=True:
[Y5F][2021-10-08 08:26:09]  load all = 1.5653s, max: 1.5911s, curr: 1.5820s, itr(B14): 0.1118s
[Y5F][2021-10-08 08:26:09]    imread = 0.0987s, max: 0.1247s, curr: 0.0940s, itr(B14): 0.0070s
[Y5F][2021-10-08 08:26:09]    resize = 0.0132s, max: 0.0362s, curr: 0.0161s, itr(B14): 0.0009s
[Y5F][2021-10-08 08:26:09]  np_batch = 0.0141s, max: 0.0170s, curr: 0.0130s, itr(B14): 0.0010s
[Y5F][2021-10-08 08:26:09]  predict_batch = 0.4262s, max: 0.4544s, curr: 0.4198s, itr(B14): 0.0304s
[Y5F][2021-10-08 08:26:09]    load_gpu = 0.0089s, max: 0.0120s, curr: 0.0061s, itr(B14): 0.0006s
[Y5F][2021-10-08 08:26:09]    normalize = 0.0000s, max: 0.0010s, curr: 0.0000s, itr(B14): 0.0000s
[Y5F][2021-10-08 08:26:09]    b14 = 0.4006s, max: 0.4324s, curr: 0.4014s, itr(B14): 0.0286s
[Y5F][2021-10-08 08:26:09]    nms = 0.0166s, max: 0.0329s, curr: 0.0123s, itr(B14): 0.0012s
[Y5F][2021-10-08 08:26:09]      to_cpu = 0.0007s, max: 0.0040s, curr: 0.0000s, itr(B14): 0.0000s


  
[Y5F][2021-10-08 08:28:29] Full process GPU memory amprent: 2.65 GB
[Y5F][2021-10-08 08:28:29] Timing results GPU_SYNC=False, TO_CPU=True:
[Y5F][2021-10-08 08:28:29]  predict_batch = 0.4269s, max: 0.4592s, curr: 0.4229s, itr(B14): 0.0305s
[Y5F][2021-10-08 08:28:29]    load_gpu = 0.0092s, max: 0.0121s, curr: 0.0100s, itr(B14): 0.0007s
[Y5F][2021-10-08 08:28:29]    normalize = 0.0000s, max: 0.0030s, curr: 0.0000s, itr(B14): 0.0000s
[Y5F][2021-10-08 08:28:29]    b14 = 0.0167s, max: 0.0423s, curr: 0.0289s, itr(B14): 0.0012s
[Y5F][2021-10-08 08:28:29]    nms = 0.4010s, max: 0.4239s, curr: 0.3840s, itr(B14): 0.0286s
[Y5F][2021-10-08 08:28:29]      to_cpu = 0.0008s, max: 0.0040s, curr: 0.0010s, itr(B14): 0.0001s



[Y5F][2021-10-08 08:31:39] Full process GPU memory amprent: 2.64 GB
[Y5F][2021-10-08 08:31:39] Timing results GPU_SYNC=False, TO_CPU=False:
[Y5F][2021-10-08 08:31:39]  predict_batch = 0.4240s, max: 0.4518s, curr: 0.4219s, itr(B14): 0.0303s
[Y5F][2021-10-08 08:31:39]    load_gpu = 0.0089s, max: 0.0144s, curr: 0.0080s, itr(B14): 0.0006s
[Y5F][2021-10-08 08:31:39]    normalize = 0.0000s, max: 0.0010s, curr: 0.0000s, itr(B14): 0.0000s
[Y5F][2021-10-08 08:31:39]    b14 = 0.0164s, max: 0.0439s, curr: 0.0290s, itr(B14): 0.0012s
[Y5F][2021-10-08 08:31:39]    nms = 0.3986s, max: 0.4184s, curr: 0.3849s, itr(B14): 0.0285s


[Y5F][2021-10-08 10:38:03] Timing results GPU_SYNC=True, TO_CPU=True, AMP=True:
[Y5F][2021-10-08 10:38:03]  predict_batch = 0.2400s, max: 0.2703s, curr: 0.2364s, itr(B14): 0.0171s
[Y5F][2021-10-08 10:38:03]    load_gpu = 0.0085s, max: 0.0123s, curr: 0.0081s, itr(B14): 0.0006s
[Y5F][2021-10-08 10:38:03]    normalize = 0.0001s, max: 0.0040s, curr: 0.0000s, itr(B14): 0.0000s
[Y5F][2021-10-08 10:38:03]    b14 = 0.2183s, max: 0.2463s, curr: 0.2163s, itr(B14): 0.0156s
[Y5F][2021-10-08 10:38:03]    nms = 0.0131s, max: 0.0354s, curr: 0.0121s, itr(B14): 0.0009s
[Y5F][2021-10-08 10:38:03]      to_cpu = 0.0008s, max: 0.0041s, curr: 0.0000s, itr(B14): 0.0001s

[Y5F][2021-10-08 13:03:52] Timing results GPU_SYNC=True, AMP=True @(14, 3, 896, 1280):
[Y5F][2021-10-08 13:03:52]  predict_batch = 0.3139s, max: 0.3326s, curr: 0.3256s, itr(B14): 0.0224s
[Y5F][2021-10-08 13:03:52]    load_gpu = 0.0115s, max: 0.0161s, curr: 0.0081s, itr(B14): 0.0008s
[Y5F][2021-10-08 13:03:52]    normalize = 0.0074s, max: 0.0142s, curr: 0.0079s, itr(B14): 0.0005s
[Y5F][2021-10-08 13:03:52]    b14 = 0.2785s, max: 0.2916s, curr: 0.2816s, itr(B14): 0.0199s
[Y5F][2021-10-08 13:03:52]    nms = 0.0166s, max: 0.0345s, curr: 0.0280s, itr(B14): 0.0012s
[Y5F][2021-10-08 13:03:52]      to_cpu = 0.0010s, max: 0.0040s, curr: 0.0000s, itr(B14): 0.0001s

[Y5F][2021-10-08 19:24:22] Full process GPU memory amprent: 3.84 GB
[Y5F][2021-10-08 19:24:22] Timing results GPU_SYNC=True, AMP=False @(14, 3, 896, 1280):
[Y5F][2021-10-08 19:24:22]  predict_batch = 0.4822s, max: 0.5201s, curr: 0.5201s, itr(B14): 0.0344s
[Y5F][2021-10-08 19:24:22]    load_gpu = 0.0112s, max: 0.0180s, curr: 0.0110s, itr(B14): 0.0008s
[Y5F][2021-10-08 19:24:22]    normalize = 0.0073s, max: 0.0143s, curr: 0.0060s, itr(B14): 0.0005s
[Y5F][2021-10-08 19:24:22]    b14 = 0.4473s, max: 0.4718s, curr: 0.4718s, itr(B14): 0.0319s
[Y5F][2021-10-08 19:24:22]    nms = 0.0165s, max: 0.0337s, curr: 0.0313s, itr(B14): 0.0012s
[Y5F][2021-10-08 19:24:22]      to_cpu = 0.0009s, max: 0.0041s, curr: 0.0031s, itr(B14): 0.0001s

[Y5F][2021-10-12 15:56:01] Full process GPU memory amprent: 3.33 GB
[Y5F][2021-10-12 15:56:01] Timing results GPU_SYNC:True, AMP:True @(14, 3, 768, 1280):
[Y5F][2021-10-12 15:56:01]  load all = 0.4279s, max: 0.4558s, curr: 0.4558s, itr(B14): 0.0306s
[Y5F][2021-10-12 15:56:01]    imread = 0.0216s, max: 0.0389s, curr: 0.0199s, itr(B14): 0.0015s
[Y5F][2021-10-12 15:56:01]    resize = 0.0080s, max: 0.0130s, curr: 0.0080s, itr(B14): 0.0006s
[Y5F][2021-10-12 15:56:01]  np_batch = 0.0150s, max: 0.0170s, curr: 0.0130s, itr(B14): 0.0011s
[Y5F][2021-10-12 15:56:01]  predict_batch = 0.2738s, max: 0.3029s, curr: 0.2665s, itr(B14): 0.0196s
[Y5F][2021-10-12 15:56:01]    load_gpu = 0.0099s, max: 0.0209s, curr: 0.0100s, itr(B14): 0.0007s
[Y5F][2021-10-12 15:56:01]    normalize = 0.0061s, max: 0.0150s, curr: 0.0065s, itr(B14): 0.0004s
[Y5F][2021-10-12 15:56:01]    b14 = 0.2418s, max: 0.2541s, curr: 0.2371s, itr(B14): 0.0173s
[Y5F][2021-10-12 15:56:01]    nms = 0.0159s, max: 0.0329s, curr: 0.0130s, itr(B14): 0.0011s
[Y5F][2021-10-12 15:56:01]      to_cpu = 0.0007s, max: 0.0011s, curr: 0.0010s, itr(B14): 0.0000s


[Y5F][2021-10-18 19:19:53] Full process GPU memory amprent: 3.15 GB
[Y5F][2021-10-18 19:19:53] Timing results 'yolov5l6' FP16 GPU_SYNC:True, AMP:True @(14, 3, 896, 1280):
[Y5F][2021-10-18 19:19:53]  load all = 0.4264s, max: 0.4279s, curr: 0.4279s, itr(B14): 0.0305s
[Y5F][2021-10-18 19:19:53]    imread = 0.0226s, max: 0.0409s, curr: 0.0199s, itr(B14): 0.0016s
[Y5F][2021-10-18 19:19:53]    resize = 0.0082s, max: 0.0110s, curr: 0.0080s, itr(B14): 0.0006s
[Y5F][2021-10-18 19:19:53]  np_batch = 0.0150s, max: 0.0150s, curr: 0.0150s, itr(B14): 0.0011s
[Y5F][2021-10-18 19:19:53]  predict_batch = 0.3080s, max: 0.3291s, curr: 0.3043s, itr(B14): 0.0220s
[Y5F][2021-10-18 19:19:53]    load_gpu = 0.0218s, max: 0.0309s, curr: 0.0203s, itr(B14): 0.0016s
[Y5F][2021-10-18 19:19:53]    normalize = 0.0045s, max: 0.0053s, curr: 0.0047s, itr(B14): 0.0003s
[Y5F][2021-10-18 19:19:53]    b14 = 0.2675s, max: 0.2832s, curr: 0.2662s, itr(B14): 0.0191s
[Y5F][2021-10-18 19:19:53]    nms = 0.0143s, max: 0.0319s, curr: 0.0130s, itr(B14): 0.0010s
[Y5F][2021-10-18 19:19:53]      to_cpu = 0.0006s, max: 0.0010s, curr: 0.0000s, itr(B14): 0.0000s


[Y5F][2021-10-18 19:23:10] Timing results 'yolov5l6' FP16 GPU_SYNC:True, AMP:False @(14, 3, 896, 1280):
[Y5F][2021-10-18 19:23:10]  load all = 0.4264s, max: 0.4279s, curr: 0.4279s, itr(B14): 0.0305s
[Y5F][2021-10-18 19:23:10]    imread = 0.0222s, max: 0.0339s, curr: 0.0199s, itr(B14): 0.0016s
[Y5F][2021-10-18 19:23:10]    resize = 0.0082s, max: 0.0090s, curr: 0.0090s, itr(B14): 0.0006s
[Y5F][2021-10-18 19:23:10]  np_batch = 0.0185s, max: 0.0219s, curr: 0.0150s, itr(B14): 0.0013s
[Y5F][2021-10-18 19:23:10]  predict_batch = 0.3081s, max: 0.3428s, curr: 0.3143s, itr(B14): 0.0220s
[Y5F][2021-10-18 19:23:10]    load_gpu = 0.0217s, max: 0.0263s, curr: 0.0258s, itr(B14): 0.0015s
[Y5F][2021-10-18 19:23:10]    normalize = 0.0045s, max: 0.0070s, curr: 0.0046s, itr(B14): 0.0003s
[Y5F][2021-10-18 19:23:10]    b14 = 0.2677s, max: 0.2870s, curr: 0.2699s, itr(B14): 0.0191s
[Y5F][2021-10-18 19:23:10]    nms = 0.0143s, max: 0.0259s, curr: 0.0140s, itr(B14): 0.0010s
[Y5F][2021-10-18 19:23:10]      to_cpu = 0.0006s, max: 0.0011s, curr: 0.0000s, itr(B14): 0.0000s



[Y5F][2021-10-18 19:25:11] Full process GPU memory amprent: 3.49 GB
[Y5F][2021-10-18 19:25:11] Timing results 'yolov5l6' FP32 GPU_SYNC:True, AMP:True @(14, 3, 896, 1280):
[Y5F][2021-10-18 19:25:11]  load all = 0.4269s, max: 0.4289s, curr: 0.4289s, itr(B14): 0.0305s
[Y5F][2021-10-18 19:25:11]    imread = 0.0235s, max: 0.0389s, curr: 0.0209s, itr(B14): 0.0017s
[Y5F][2021-10-18 19:25:11]    resize = 0.0092s, max: 0.0279s, curr: 0.0080s, itr(B14): 0.0007s
[Y5F][2021-10-18 19:25:11]  np_batch = 0.0214s, max: 0.0279s, curr: 0.0150s, itr(B14): 0.0015s
[Y5F][2021-10-18 19:25:11]  predict_batch = 0.3368s, max: 0.3680s, curr: 0.3312s, itr(B14): 0.0241s
[Y5F][2021-10-18 19:25:11]    load_gpu = 0.0418s, max: 0.0539s, curr: 0.0439s, itr(B14): 0.0030s
[Y5F][2021-10-18 19:25:11]    normalize = 0.0073s, max: 0.0083s, curr: 0.0056s, itr(B14): 0.0005s
[Y5F][2021-10-18 19:25:11]    b14 = 0.2737s, max: 0.2925s, curr: 0.2687s, itr(B14): 0.0195s
[Y5F][2021-10-18 19:25:11]    nms = 0.0140s, max: 0.0259s, curr: 0.0130s, itr(B14): 0.0010s
[Y5F][2021-10-18 19:25:11]      to_cpu = 0.0006s, max: 0.0011s, curr: 0.0010s, itr(B14): 0.0000s

[Y5F][2021-10-18 19:27:56] Full process GPU memory amprent: 3.49 GB
[Y5F][2021-10-18 19:27:56] Timing results 'yolov5l6' FP32 GPU_SYNC:True, AMP:True @(14, 3, 896, 1280):
[Y5F][2021-10-18 19:27:56]  load all = 0.4264s, max: 0.4289s, curr: 0.4289s, itr(B14): 0.0305s
[Y5F][2021-10-18 19:27:56]    imread = 0.0223s, max: 0.0279s, curr: 0.0199s, itr(B14): 0.0016s
[Y5F][2021-10-18 19:27:56]    resize = 0.0088s, max: 0.0279s, curr: 0.0090s, itr(B14): 0.0006s
[Y5F][2021-10-18 19:27:56]  np_batch = 0.0155s, max: 0.0160s, curr: 0.0150s, itr(B14): 0.0011s
[Y5F][2021-10-18 19:27:56]  predict_batch = 0.3363s, max: 0.3551s, curr: 0.3311s, itr(B14): 0.0240s
[Y5F][2021-10-18 19:27:56]    load_gpu = 0.0413s, max: 0.0519s, curr: 0.0394s, itr(B14): 0.0029s
[Y5F][2021-10-18 19:27:56]    normalize = 0.0073s, max: 0.0121s, curr: 0.0080s, itr(B14): 0.0005s
[Y5F][2021-10-18 19:27:56]    b14 = 0.2735s, max: 0.2862s, curr: 0.2708s, itr(B14): 0.0195s
[Y5F][2021-10-18 19:27:56]    nms = 0.0142s, max: 0.0309s, curr: 0.0130s, itr(B14): 0.0010s
[Y5F][2021-10-18 19:27:56]      to_cpu = 0.0007s, max: 0.0020s, curr: 0.0000s, itr(B14): 0.0001s

"""
from decentra_vision.draw_utils import DrawUtils
from naeural_core import Logger
import numpy as np
import torch as th
import cv2
import os

import sys
from pathlib import Path

LOAD_HUB = False


if LOAD_HUB:
  FILE = Path('utils/y5/models/yolo.py').resolve()
  ROOT = FILE.parents[1]  # YOLOv5 root directory
  if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
  # ROOT = ROOT.relative_to(Path.cwd())  # relative
  from models.experimental import attempt_load
  from y5utils.augmentations import letterbox
  from y5utils.general import non_max_suppression, scale_coords, set_logging, check_yaml
  from models.yolo import Model
else:
  from plugins.serving.architectures.y5.yolo import Model
  from plugins.serving.architectures.y5.augmentations import letterbox
  from plugins.serving.architectures.y5.general import non_max_suppression, scale_coords, set_logging, check_yaml
  YAML_PATH = 'inference/architectures/y5/configs'


_D1 = [
  'xperimental/_images/FPs/686097.png',
  'xperimental/_images/FPs/686102.png',
  'xperimental/_images/FPs/700006.png',
  'xperimental/_images/FPs/699720.png',
  'xperimental/_images/FPs/688220.png',
]

_D2 = [
  ]

_MODELS = {
    'yolov5l6': 'https://www.dropbox.com/s/1o4ad4y93w9a93j/yolov5l6.pt?dl=1',
    'yolov5x6': 'https://www.dropbox.com/s/sc2i1qpuj5te0lg/yolov5x6.pt?dl=1',
    'yolov5s6': 'https://www.dropbox.com/s/yj4gnqugsp8tpzv/yolov5s6.pt?dl=1',
}


_WEIGHTS = {
    'yolov5l6': 'https://www.dropbox.com/s/kciushsvber05y5/yolov5l6_weights.pt?dl=1',
    'yolov5x6': 'https://www.dropbox.com/s/n1wr3sysk3zaglu/yolov5x6_weights.pt?dl=1',
    'yolov5s6': 'https://www.dropbox.com/s/98ct9x0ckaprzsu/yolov5s6_weights.pt?dl=1',
}

_COCO = 'https://www.dropbox.com/s/krzygbdl6qkzf8e/coco.txt?dl=1'


def load_yolo(log, model_name, dev, fused=True):
  fn_weights_base = model_name + '_weights.pt'
  url = _WEIGHTS[model_name]
  l.maybe_download_model(url, fn_weights_base)
  l.maybe_download_model(_COCO, 'coco_classes.txt')
  fn_weights = l.get_models_file(fn_weights_base)
  fn_def = os.path.join(YAML_PATH, model_name + '.yaml')
  if fn_weights is None:
    log.P("Cant find weights file '{}'. Attempting download".format(fn))
  if fn_def is None:
    raise ValueError("Cant find definition file '{}'".format(fn))

  cfg = check_yaml(fn_def)  # check YAML
  log.P("Creating model from '{}'".format(cfg))
  model = Model(cfg)
  model.float()
  model.eval()
  if fused:
    model.fuse()
  log.P("Loading weights from '{}'".format(fn_weights))
  model.load_state_dict(th.load(fn_weights))
  model.to(dev)
  return model


def th_predict(log, model, np_inputs, gpu_sync, use_amp, fp16, dev):
  th_dtype = th.float16 if fp16 else th.float32
  np_dtype = 'float16' if fp16 else 'float32'
  np_inputs = np_inputs.astype(np_dtype)
  with th.cuda.amp.autocast(enabled=use_amp):
    with th.no_grad():
      log.start_timer('predict_batch')

      log.start_timer('load_gpu')
      th_x = th.tensor(np_inputs, device=dev, dtype=th_dtype)
      if gpu_sync:
        th.cuda.synchronize()
      log.stop_timer('load_gpu')

      log.start_timer('normalize')
      th_x = th_x / 255.
      log.stop_timer('normalize')

      log.start_timer('b14')
      res = model(th_x)
      preds = res[0]
      if gpu_sync:
        th.cuda.synchronize()
      log.stop_timer('b14')

      log.start_timer('nms')
      pred_nms = non_max_suppression(
        prediction=preds,
        conf_thres=0.25,
        iou_thres=0.45,
        classes=None,
        agnostic=False,
        max_det=300
      )

      if gpu_sync:
        th.cuda.synchronize()

      log.start_timer('to_cpu')
      cpu_pred_nms = [x.cpu().numpy() for x in pred_nms]
      log.stop_timer('to_cpu')

      log.stop_timer('nms')

      log.stop_timer('predict_batch')

      return cpu_pred_nms


if __name__ == '__main__':
  l = Logger('Y5F', base_folder='.', app_folder='_cache', TF_KERAS=False)
  l.gpu_info()

  painter = DrawUtils(log=l)

  RUNS_IMG = 3  # runs to determine resize speed
  RUNS_INF = 10  # runs to determine inference speed
  BATCH_SIZE = 14
  GPU_SYNC = True
  AMP = False
  _BASE = _D1  # images to use
  SHOW = False
  MODEL = 'yolov5l6'
  FP16 = True
  # AMP = AMP if not FP16 else False

  NR_SHOW = len(_BASE)
  fns = (_BASE * 20)[:BATCH_SIZE]
  url = _MODELS[MODEL]
  fn_base = MODEL + '.pt'
  INPUT_SIZE = (896, 1280)
  AUTO = len(INPUT_SIZE) == 1

  g1 = l.gpu_info()[0]
  l.P("Pre-load model GPU free: {:.1f} GB".format(g1['FREE_MEM']))
  selected_dev = th.device('cuda')

  if LOAD_HUB:
    l.P("Preparing model '{}' using TorchHub approach".format(
      MODEL), color='green')
    model_type = MODEL
    l.maybe_download_model(url, fn_base)
    fn_model = l.get_models_file(fn_base)
    set_logging()
    model = attempt_load(fn_model, map_location=selected_dev)
    fn_weights = os.path.join(l.get_models_folder(), MODEL + '_weights.pt')
    if l.get_models_file(fn_weights) is None:
      th.save(model.state_dict(), fn_weights)
    class_names = model.module.names if hasattr(model, 'module') else model.names
    l.save_pickle_to_models(class_names, 'coco_class_names.pkl')
  else:
    model_type = MODEL + '_W'
    l.P("Preparing model '{}' using simplified approach".format(
      MODEL), color='green')
    model = load_yolo(
      log=l,
      model_name=MODEL,
      dev=selected_dev
    )
    class_names = l.load_models_json('coco_classes.txt')
    if class_names is None:
      raise ValueError('Could not find class names')

  g2 = l.gpu_info()[0]
  l.P("Post-load model GPU free: {:.1f} GB => initial amprent {:.1f} GB".format(
    g2['FREE_MEM'], g1['FREE_MEM'] - g2['FREE_MEM']))
  stride = model.stride.max().cpu().item()
  imgs_bgr = []
  l.P("Loading images...")
  for _ in range(RUNS_IMG):
    l.start_timer('load all')
    imgs = []
    for fn in fns:
      l.start_timer('imread')
      img = cv2.imread(fn)
      if img is None:
        raise ValueError('Error reading {}'.format(fn))
      imgs_bgr.append(img)
      l.stop_timer('imread')
      l.start_timer('resize')
      img_resized = letterbox(
        img,
        new_shape=INPUT_SIZE,
        stride=stride,
        auto=AUTO,
      )[0]
      img_rgb = img_resized[:, :, ::-1]
      img_th = np.transpose(img_rgb, [2, 0, 1])
      np_img = np.ascontiguousarray(img_th)
      l.stop_timer('resize')
      imgs.append(np_img)
    l.stop_timer('load all')

  l.P("To batch...")
  t_np = []
  for _ in range(RUNS_IMG):
    l.start_timer("np_batch")
    np_imgs = np.array(imgs)
    t_np.append(l.stop_timer("np_batch"))
  input_size = np_imgs.shape[-2:]
  if len(np_imgs.shape) == 3:
    np_imgs = np.expand_dims(np_imgs, 0)

  model.eval()
  if FP16:
    model.half()

  model_dtype = next(model.parameters()).dtype
  l.P("Model is using '{}'".format(model_dtype))
  t_upl = []
  t_inf = []
  t_nms = []
  t_pre = []

  l.P("Warm-up section...")
  for _ in range(2):
    th_predict(
      log=l,
      model=model,
      np_inputs=np_imgs,
      gpu_sync=GPU_SYNC,
      use_amp=AMP,
      fp16=FP16,
      dev=selected_dev,
    )

  g3 = l.gpu_info()[0]
  l.P("Post-warming model GPU free: {:.1f} GB => amprent {:.2f} GB".format(
    g3['FREE_MEM'], g2['FREE_MEM'] - g3['FREE_MEM']))

  title = "'{}' {} GPU_SYNC:{}, AMP:{} @{}".format(
    MODEL, "FP16" if FP16 else "FP32",
    GPU_SYNC, AMP, np_imgs.shape)
  l.P("Starting loop inference with {}...".format(title))
  for itr in range(RUNS_INF):
    print("\r {:.1f}%".format((itr + 1) / RUNS_INF * 100), end='', flush=True)
    np_imgs2 = np_imgs.copy()
    np.random.shuffle(np_imgs2)
    th.cuda.empty_cache()

    th_predict(
      log=l,
      model=model,
      np_inputs=np_imgs2,
      gpu_sync=GPU_SYNC,
      use_amp=AMP,
      fp16=FP16,
      dev=selected_dev,
    )

  g4 = l.gpu_info()[0]
  l.P("Post-loop GPU free: {:.1f} GB => amprent {:.2f} GB".format(
    g4['FREE_MEM'], g3['FREE_MEM'] - g4['FREE_MEM']))

  l.P("Full process GPU memory amprent: {:.2f} GB".format(g1['FREE_MEM'] - g4['FREE_MEM']), color='y')

  l.show_timers(
    title=title,
    div=np_imgs.shape[0])

  l.P("Last inference...")
  cpu_preds_nms = th_predict(
    log=l,
    model=model,
    np_inputs=np_imgs,
    gpu_sync=GPU_SYNC,
    use_amp=AMP,
    dev=selected_dev,
    fp16=FP16,
  )
  for i in range(NR_SHOW):
    np_pred = cpu_preds_nms[i]
    im0 = imgs_bgr[i]
    np_pred[:, :4] = scale_coords(input_size, np_pred[:, :4], im0.shape).round()
    lst_inf = []
    for det in np_pred:
      L, T, R, B, P, C = det  # order is [left, top, right, proba, class]
      dct_obj = {
        'TLBR_POS': [T, L, B, R],
        'PROB_PRC': P,
        'TYPE': class_names[int(C)],
      }
      lst_inf.append(dct_obj)
    img_bgr = painter.draw_inference_boxes(
      image=im0,
      lst_inf=lst_inf,
    )
    img_bgr = painter.resize(img_bgr, 1000, int(900 / img_bgr.shape[0] * img_bgr.shape[1]))
    if SHOW:
      painter.show(fns[i], img_bgr)
    fn = os.path.split(fns[i])[1][:-4]
    fn = '{}_{}_{}_{}_A{}_{}.png'.format(
      fn, model_type, *input_size, int(AMP),
      'FP16' if FP16 else 'FP32'
    )
    l.P("  Saving '{}'".format(fn))
    painter.save(img_bgr, fn=fn)



import numpy as np
import os

from ratio1 import BaseDecentrAIObject

from transformers import TFBertModel, BertTokenizer


__VER__ = '0.1.0.7'
__DEFAULT_MODEL__ = '_allan_data/_ro_bert/20200520'


class RomBERT(BaseDecentrAIObject):
  """
  Romanian BERT (or any BERT) wrapper
  """  
  def __init__(self, 
               max_sent=None,
               model_folder=None, 
               DEBUG=True,
               **kwargs):
    """
    
    Parameters
    ----------
    max_sent : int, optional
      If the downstream model needs a fixed len of seq here you must put it. Otherwise,
      the engine will pad at max len. The default is None.
    model_folder : str, optional
      Where to get the BERT weights from. The default is None.
    DEBUG : bool, optional
      if you need full debug. The default is True.
    **kwargs : TYPE
      other kwargs.

    Raises
    ------
    ValueError
      DESCRIPTION.

    Returns
    -------
    None.

    """    
    self.__version__ = __VER__
    super().__init__(DEBUG=DEBUG,**kwargs)    
    if model_folder is None:
      self._model_folder = os.path.join(
        self.log._get_cloud_base_folder('dropbox'),
        __DEFAULT_MODEL__
        )
    else:
      self.model_folder= model_folder
    if not os.path.isdir(self._model_folder):
      raise ValueError("{} does not exists".format(self._model_folder))
    self.max_sent = max_sent
    self._load_models()
    return


  def _load_models(self):
    self.log.start_timer('load_models')
    self.P("Loading tokenizer...")
    self.tokeng = BertTokenizer.from_pretrained(self._model_folder)
    self.P("Loading pretrained tf BERT...")
    self.embeng = TFBertModel.from_pretrained(self._model_folder)
    elapsed = self.log.stop_timer('load_models')
    self.P("Done loading models in {:.1f}s.".format(elapsed))
    self.P("Loaded model {}('{}') with {:,.2f} Mil params".format(
        self.embeng.__class__.__name__,
        self.embeng.base_model_prefix,
        self.embeng.num_parameters() / (1000**2),
        ),
      color='g'
      )
    return
  
  
  def text2embeds(self, sents, return_only_summary=False):
    """
    

    Parameters
    ----------
    sent : str or list[str]
      sentence or list of N sentences 
      
    return_only_summary: bool, default False
      return only the embedding "summary" consisting of the '[CLS]' token embedding

    Returns
    -------
    embeds : ndarray (N, E)
      returns the embeddings

    """
    self.D("Running {} text2embeds with return_only_summary={}".format(
      self.__class__.__name__,
      return_only_summary
      ))
    self.log.start_timer('text2embeds')
    self.np_last_ids, self.np_last_mask = self.text2ids(
      sents=sents,
      skip_pad=False,
      max_sent=self.max_sent
      )
    self.np_last_embeds, self.np_last_clf = self.embeng.predict([
      self.np_last_ids, 
      self.np_last_mask
      ])
    elapsed = self.log.stop_timer('text2embeds')
    self.D("  Done text2embeds in {:.1f}s.".format(elapsed))
    if not return_only_summary:
      return self.np_last_embeds
    else:
      return self.np_last_clf
  
  
  def text2tokens(self, sent, skip_pad=True):
    """
    

    Parameters
    ----------
    sent : str or list[str]
      sentence or list of N sentences 
      
    skip_pad:
      skip the padding

    Returns
    -------
    output : list
      list of word piece tokenization

    """
    self.log.start_timer('text2embeds')
    np_ids = self.text2ids(sent, return_mask=False, skip_pad=skip_pad)
    output = []
    for ids in np_ids:
      tokens = self.tokeng.convert_ids_to_tokens(ids)    
      output.append(tokens)
    self.log.stop_timer('text2embeds')
    return output
  
  def text2ids(self, sents, return_mask=True, skip_pad=False, max_sent=None):
    """
    

    Parameters
    ----------
    sents : str or list[str]
      sentence or list of N sentences 
    return_mask : bool, optional 
      If attention mask must be returned (derived from padding). The default is True.
    skip_pad : bool, optional 
      Skip padding if we just need the ids. The default is False.

    Returns
    -------
      the list or ndarray of ids

    """
    if type(sents) == str:
      sents = [sents]
    elif type(sents) == list:
      assert type(sents[0]) == str, "input must be `str` or `list[str]`"
    self.log.start_timer('text2ids')
    if max_sent or not skip_pad:
      data = self.tokeng.batch_encode_plus(
          sents,
          add_special_tokens=True,
          return_attention_mask=True,
          pad_to_max_length=True,
          max_length=max_sent,
          )
    else:
      data = self.tokeng.batch_encode_plus(
          sents,
          add_special_tokens=True,
          return_attention_mask=True,
          )
    np_ids = np.array(data['input_ids'])
    np_mask = np.array(data['attention_mask'])
    self.log.start_timer('text2ids')
    if return_mask:
      return np_ids, np_mask
    else:
      return np_ids

      
  
if __name__ == '__main__':
  from naeural_core import Logger

  l = Logger(lib_name='ALBERT', config_file='tagger/brain/configs/config_cv_test.txt')
  
  if 'eng' not in globals():
    eng = RomBERT(log=l)
  
  lines = ['ana are mere.','mara are multe pere!']
  one_sent = 'ana are mere fara punct'
  corpus = [
    'Totul este supercariflagiristicus', 
    'Ma duc sa imi iau un Samsung ca am luat leafa si sunt bogat!',
    'Iphone 8 husa rosie, trusa machiaj avon',
    'Scutece librese, camasa noapte, pantofi jack wolfskin',
    ]
  
  e1 = eng.text2embeds(lines)
  l.P("Generated {} embeds out of {}".format(e1.shape, lines))

  e2 = eng.text2embeds(lines, return_only_summary=True)
  l.P("Generated {} embeds out of {}".format(e2.shape, lines))  
  
  # l.P("e1[:,0] == e2 ({})".format(np.allclose(e1[:,0],e2)))

  l.P("Tokenizer analysis:")
  l.P('  {} = {}'.format(one_sent, eng.text2tokens(one_sent)))
  l.P('  {} = {}'.format(lines, eng.text2tokens(lines)))
  t = eng.text2tokens(corpus)
  l.P("  Full corpus tokens:")
  for _line in t:
    print(_line)
  



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Callback utils
"""


class Callbacks:
    """"
    Handles all registered callbacks for YOLOv5 Hooks
    """

    # Define the available callbacks
    _callbacks = {
        'on_pretrain_routine_start': [],
        'on_pretrain_routine_end': [],

        'on_train_start': [],
        'on_train_epoch_start': [],
        'on_train_batch_start': [],
        'optimizer_step': [],
        'on_before_zero_grad': [],
        'on_train_batch_end': [],
        'on_train_epoch_end': [],

        'on_val_start': [],
        'on_val_batch_start': [],
        'on_val_image_end': [],
        'on_val_batch_end': [],
        'on_val_end': [],

        'on_fit_epoch_end': [],  # fit = train + val
        'on_model_save': [],
        'on_train_end': [],

        'teardown': [],
    }

    def register_action(self, hook, name='', callback=None):
        """
        Register a new action to a callback hook

        Args:
            hook        The callback hook name to register the action to
            name        The name of the action for later reference
            callback    The callback to fire
        """
        assert hook in self._callbacks, f"hook '{hook}' not found in callbacks {self._callbacks}"
        assert callable(callback), f"callback '{callback}' is not callable"
        self._callbacks[hook].append({'name': name, 'callback': callback})

    def get_registered_actions(self, hook=None):
        """"
        Returns all the registered actions by callback hook

        Args:
            hook The name of the hook to check, defaults to all
        """
        if hook:
            return self._callbacks[hook]
        else:
            return self._callbacks

    def run(self, hook, *args, **kwargs):
        """
        Loop through the registered actions and fire all callbacks

        Args:
            hook The name of the hook to check, defaults to all
            args Arguments to receive from YOLOv5
            kwargs Keyword Arguments to receive from YOLOv5
        """

        assert hook in self._callbacks, f"hook '{hook}' not found in callbacks {self._callbacks}"

        for logger in self._callbacks[hook]:
            logger['callback'](*args, **kwargs)



"""

1. Delete Manager
2. Open multiprocessing SharedMemory (serving_id = 'Slave2) & alloc in master 100 images
3. Write to SharedMemory & send to Slave nr of images
4. Slave gets nr of images and retrieves from SharedMemory

"""

import multiprocessing
import numpy as np
import time

from uuid import uuid4

from naeural_core.xperimental.manager_shm.slave import Slave1, Slave2  # Ensure this import matches your file organization
from naeural_core.utils.system_shared_memory import NumpySharedMemory
from naeural_core import Logger
from math import prod


MSG_SHAPE = (24, 3, 720, 1280)
MSG_SIZE = prod(MSG_SHAPE) * np.dtype('uint8').itemsize


class Master:
  """
  Master class to manage slave processes and communication, including sending specific payloads
  and receiving acknowledgments with counters.
  """
  def __init__(self, log, use_shared_memory=True):
    # self.manager = multiprocessing.Manager()
    self.queue1_snd = multiprocessing.Queue()
    self.queue1_rcv = multiprocessing.Queue()
    
    self.queue2_snd = multiprocessing.Queue()
    self.queue2_rcv = multiprocessing.Queue()
    # self.dict2 = self.manager.dict()
    self.shared_mem_kwargs = {
      'mem_name' : 'Slave2',
      'mem_size' : MSG_SIZE,
      'np_shape' : MSG_SHAPE,
      'np_type' : 'uint8',
      'log' : log,
    }
    self.dict2 = NumpySharedMemory(**self.shared_mem_kwargs, create=True)
    self.__counter = 0
    self.use_shared_memory = use_shared_memory

    self.slave1 = Slave1(self.queue1_snd, self.queue1_rcv)
    self.slave2 = Slave2(self.queue2_snd, self.queue2_rcv, {**self.shared_mem_kwargs, 'log': None}, use_shared_memory=use_shared_memory)
    return

  def generate_payload(self, nr_images=24, image_size=(3, 720, 1280)):
    """
    Generates a list of numpy arrays with specified values based on a counter.


    Returns:
    - list: A list of numpy arrays.
    """
    payload = {
      'DATA' : [],
      'COUNTER' : 0,
    }
    for i in range(nr_images):
      self.__counter += 1
      payload['DATA'].append(
        np.ones(image_size, dtype='uint8') * self.__counter
      )
    payload['COUNTER'] = self.__counter
    return payload
    

  def run_slaves_and_send_payload(self, iterations=10):
    """
    Starts slave processes, sends payloads with counters, and measures round-trip time.
    """
    try:
      self.slave1.start()
      self.slave2.start()
    except Exception as e:
      a = 1
      print(f"Error starting slave processes: {e}", flush=True)
      return
    slave1_timers = []
    slave2_timers = []
    slave21 = []
    
    USER_ARRAYS = True
    USE_SINGLE_BUFFER = True
    
    SHM_WORKER_ID = 'Slave2'
    
    if USER_ARRAYS and not self.use_shared_memory:
      self.dict2[SHM_WORKER_ID] = np.zeros((24, 3, 720, 1280), dtype='uint8')

    for counter in range(1, iterations):  # Example: Send 5 payloads
      payload = self.generate_payload()
      counter = payload['COUNTER']
      print(f"Master: Sending payload with last counter: {counter}...", flush=True)
      
      # Send payload to Slave1 and measure round-trip time
      
      start_time1 = time.time()
      self.queue1_snd.put(payload)
      recv1 = self.queue1_rcv.get()  # Receive counter acknowledgment
      elapsed1 = time.time() - start_time1
      slave1_timers.append(elapsed1)
      print(f"Master: Slave1 ack payload: {recv1} in {elapsed1:.4f} seconds.", flush=True)

      # Send payload to Slave2 and measure round-trip time
      start_time2 = time.time()
      if USE_SINGLE_BUFFER:
        pointer = SHM_WORKER_ID
      else:
        pointer = str(uuid4())[:5]
      imgs = np.array(payload['DATA'])
      t21 = time.time()
      if self.use_shared_memory:
        self.dict2.write(np.array(payload['DATA']))
      else:
        if not USER_ARRAYS:
          self.dict2[pointer] = payload['DATA']
        else:
          self.dict2[pointer][:,:,:,:] = imgs[:,:,:,:]
      e21 = time.time() - t21
      self.queue2_snd.put({'COUNTER' : pointer, 'DATA' :  None, 'IS_ARRAY' : USER_ARRAYS})  # Signal to process data
      recv2 = self.queue2_rcv.get()  # Receive counter acknowledgment
      elapsed2 = time.time() - start_time2
      slave2_timers.append(elapsed2)
      slave21.append(e21)
      data = self.dict2[pointer] if not self.use_shared_memory else self.dict2.read()
      print(f"Master: Slave2 ack payload: {recv2} in {elapsed2:.4f} seconds for {id(data)}", flush=True)
      if not USE_SINGLE_BUFFER:
        del self.dict2[pointer]

    # Cleanup
    print(f"Results:\n Slave1 timers: {np.mean(slave1_timers):.4f} seconds\n Slave2 timers: {np.mean(slave2_timers):.4f} seconds  {np.mean(slave21):.4f}", flush=True)
    self.queue1_snd.put({'DATA' : [], 'COUNTER' : None,})  # Signal to Slave1 to terminate
    self.queue2_snd.put({'DATA' : [], 'COUNTER' : None,})  # Signal to Slave1 to terminate
    self.slave1.join()
    self.slave2.join()
    return


def test_npy_copy(n_iter=1000):
  """
  Test the time it takes to copy a numpy array with different methods.
  """
  import math

  shape = (100, 720, 1280, 3)
  data = np.random.randint(low=0, high=255, size=shape, dtype='uint8')
  timings = []
  ok = True
  x = np.zeros(shape, dtype='uint8')
  print(f'Method 1: plain : + copy')
  for _ in range(n_iter):
    start_time = time.time()
    x[:] = data[:]
    y = x.copy()
    elapsed = time.time() - start_time
    timings.append(elapsed)
    ok = ok and np.all(data == y)
  print(f"[{ok}]Copy time: {np.mean(timings):.4f} seconds", flush=True)


  timings1 = []
  ok1 = True
  x = np.zeros(math.prod(shape), dtype='uint8')
  print(f'Method 2: ravel + : + reshape + copy')
  for _ in range(n_iter):
    start_time = time.time()
    x[:] = data.ravel()[:]
    y = x.reshape(shape)
    z = y.copy()
    elapsed = time.time() - start_time
    timings1.append(elapsed)
    ok1 = ok1 and np.all(data == z)
  print(f"[{ok1}]Copy time: {np.mean(timings1):.4f} seconds", flush=True)


  timings2 = []
  ok2 = True
  x = np.zeros(math.prod(shape), dtype='uint8')
  print(f'Method 3: reshape(-1) + : + reshape + copy')
  for _ in range(n_iter):
    start_time = time.time()
    x[:] = data.reshape(-1)[:]
    y = x.reshape(shape)
    z = y.copy()
    elapsed = time.time() - start_time
    timings2.append(elapsed)
    ok2 = ok2 and np.all(data == z)
  print(f"[{ok2}]Copy time: {np.mean(timings2):.4f} seconds", flush=True)

  timings3 = []
  ok3 = True
  x = np.zeros(math.prod(shape), dtype='uint8')
  print(f'Method 4: reshape(-1) + np.contigousarray + : + reshape + copy')
  for _ in range(n_iter):
    start_time = time.time()
    x[:] = np.ascontiguousarray(data.reshape(-1))[:]
    y = x.reshape(shape)
    z = y.copy()
    elapsed = time.time() - start_time
    timings3.append(elapsed)
    ok3 = ok3 and np.all(data == z)
  print(f"[{ok3}]Copy time: {np.mean(timings3):.4f} seconds", flush=True)

  timings4 = []
  ok4 = True
  x = np.zeros(math.prod(shape), dtype='uint8')
  print(f'Method 5: reshape(-1) + : + reshape + copy')
  for _ in range(n_iter):
    start_time = time.time()
    x[:] = data.reshape(-1)[:]
    y = x.reshape(shape)
    z = y.copy()
    elapsed = time.time() - start_time
    timings4.append(elapsed)
    ok4 = ok4 and np.all(data == z)
  print(f"[{ok4}]Copy time: {np.mean(timings4):.4f} seconds", flush=True)
  return


if __name__ == "__main__":
  test_npy_copy(100)
  exit(-1)
  log = Logger(
    'SHMT', base_folder='.', app_folder='_cache'
  )
  master = Master(log=log)
  master.run_slaves_and_send_payload()



from typing import List, Tuple, Union, Any

import torch as th
import numpy as np
from naeural_core.local_libraries.nn.th.training.data.base import BaseDataLoaderFactory
from naeural_core.local_libraries.nn.th.training_utils import read_image
from naeural_core.local_libraries.nn.th.image_dataset_stage_preprocesser import PreprocessResizeWithPad, PreprocessMinMaxNorm

class SimpleImageClassifierDataLoaderFactory(BaseDataLoaderFactory):

  def __init__(self, image_height, image_width, **kwargs):
    self._image_height = image_height
    self._image_width = image_width
    super(SimpleImageClassifierDataLoaderFactory, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    return

  def _lst_on_load_preprocess(self) -> List[Tuple[Any, dict]]:
    return [
      (PreprocessResizeWithPad, dict(h=self._image_height, w=self._image_width, normalize=False))
    ]

  def _lst_right_before_forward_preprocess(self) -> List[Tuple[Any, dict]]:
    return [
      (PreprocessMinMaxNorm, dict(min_val=0, max_val=255, target_min_val=0, target_max_val=1))
    ]

  def _get_not_loaded_observations_and_labels(self) -> Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]:
    observations = self.dataset_info['paths']
    labels = self.dataset_info['path_to_idpath'][:, 0]
    return observations, labels

  def _load_x_and_y(self, observations, labels, idx) -> Tuple[Union[th.tensor, np.ndarray], Union[th.tensor, np.ndarray]]:
    path_to_img = observations[idx]
    lbl = labels[idx]
    np_img = read_image(path_to_img)
    np_lbl = np.array([lbl])
    return np_img, np_lbl

  def _preprocess(self, x, y, transform):
    x = transform(x)
    return x, y



import sys

from naeural_core import Logger
from naeural_core.utils.sys_mon import SystemMonitor


if __name__ == '__main__':
  l = Logger('PMON', base_folder='.', app_folder='_cache')
  pid = None
  str_pid = ''
  if len(sys.argv) >= 2:
    str_pid = sys.argv[1]
    try:
      pid = int(str_pid)
    except:
      pass
  l.P("Analysing {}...".format('given pid {}'.format(pid) if str_pid != '' else 'default pid'))
  sys_mon = SystemMonitor(
    monitored_prefix='',
    log=l,
  )
  
  sys_mon.log_status(full_log=False)
  


import os
import numpy as np

from datetime import datetime as dt

class _AdvancedTFKerasMixin(object):
  """
  Mixin for advanced tf keras functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_BasicTFKerasMixin`:
    - self.load_keras_model
    - self.save_keras_model

  * Obs: This mixin uses also attributes/methods of `_PickleSerializationMixin`:
    - self.load_pickle_from_models
  """

  def __init__(self):
    super(_AdvancedTFKerasMixin, self).__init__()

    try:
      from .basic_tfkeras_mixin import _BasicTFKerasMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _AdvancedTFKerasMixin without having _BasicTFKerasMixin")

    try:
      from ratio1.logging.logger_mixins.pickle_serialization_mixin import _PickleSerializationMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _AdvancedTFKerasMixin without having _PickleSerializationMixin")


    return

  @staticmethod
  def _create_layer(layer):
    import tensorflow.compat.v1 as tf
    ltype = type(layer)
    if ltype == tf.keras.layers.CuDNNLSTM:
      new_layer = tf.keras.layers.LSTM(
        units=layer.units,
        return_sequences=layer.return_sequences,
        return_state=layer.return_state,
        stateful=layer.stateful,
        name='cpu_' + layer.name,

        recurrent_activation='sigmoid',  # must!
      )
    elif ltype == tf.keras.layers.CuDNNGRU:
      new_layer = tf.keras.layers.GRU(
        units=layer.units,
        return_sequences=layer.return_sequences,
        return_state=layer.return_state,
        stateful=layer.stateful,
        name='cpu_' + layer.name,

        reset_after=True,  # must!
        recurrent_activation='sigmoid',  # must!
      )
    else:
      new_layer = layer.__class__.from_config(layer.get_config())
    return new_layer

  def model_cuda_to_cpu(self, model):
    import tensorflow.compat.v1 as tf
    new_model = tf.keras.models.clone_model(model, clone_function=_AdvancedTFKerasMixin._create_layer)
    cpus = 0
    self.P("  Converting model {} with {} layers...".format(model.name, len(model.layers)))
    for i, layer in enumerate(model.layers):
      new_layer = new_model.layers[i]
      ltype = type(layer)
      weights = layer.get_weights()
      if len(weights) > 0:
        if ltype in [tf.keras.layers.CuDNNLSTM, tf.keras.layers.CuDNNGRU]:
          new_weights = _AdvancedTFKerasMixin._convert_rnn_weights(new_layer, weights=weights)
          cpus += 1
        else:
          new_weights = weights
        new_layer.set_weights(new_weights)
    self.P("  Model {} saved with {} GPU to CPU conversions".format(model.name, cpus))
    return new_model

  def models_file_cuda_to_cpu(self, model_file):
    model_name = os.path.splitext(model_file)[0]
    model = self.load_keras_model(model_file)
    if model is not None:
      new_model = self.model_cuda_to_cpu(model)
      self.save_keras_model(new_model, label=model_name + '_cpu.h5')
    else:
      self.P("Model {} could not be loaded. Aborting conversion.".format(model_file))

  @staticmethod
  def _convert_rnn_weights(layer, weights):
    """Converts weights for RNN layers between native and CuDNN format.
    Input kernels for each gate are transposed and converted between Fortran
    and C layout, recurrent kernels are transposed. For LSTM biases are summed/
    split in half, for GRU biases are reshaped.
    Weights can be converted in both directions between `LSTM` and`CuDNNSLTM`
    and between `CuDNNGRU` and `GRU(reset_after=True)`. Default `GRU` is not
    compatible with `CuDNNGRU`.
    For missing biases in `LSTM`/`GRU` (`use_bias=False`),
    no conversion is made.
    # Arguments
        layer: Target layer instance.
        weights: List of source weights values (input kernels, recurrent
            kernels, [biases]) (Numpy arrays).
    # Returns
        A list of converted weights values (Numpy arrays).
    # Raises
        ValueError: for incompatible GRU layer/weights or incompatible biases
    """

    def transform_kernels(kernels, func, n_gates):
      """Transforms kernel for each gate separately using given function.
      # Arguments
          kernels: Stacked array of kernels for individual gates.
          func: Function applied to kernel of each gate.
          n_gates: Number of gates (4 for LSTM, 3 for GRU).
      # Returns
          Stacked array of transformed kernels.
      """
      return np.hstack([func(k) for k in np.hsplit(kernels, n_gates)])

    def transpose_input(from_cudnn):
      """Makes a function that transforms input kernels from/to CuDNN format.
      It keeps the shape, but changes between the layout (Fortran/C). Eg.:
      ```
      Keras                 CuDNN
      [[0, 1, 2],  <--->  [[0, 2, 4],
       [3, 4, 5]]          [1, 3, 5]]
      ```
      It can be passed to `transform_kernels()`.
      # Arguments
          from_cudnn: `True` if source weights are in CuDNN format, `False`
              if they're in plain Keras format.
      # Returns
          Function that converts input kernel to the other format.
      """
      order = 'F' if from_cudnn else 'C'

      def transform(kernel):
        return kernel.T.reshape(kernel.shape, order=order)

      return transform

    target_class = layer.__class__.__name__

    # convert the weights between CuDNNLSTM and LSTM
    if target_class in ['LSTM', 'CuDNNLSTM'] and len(weights) == 3:
      # determine if we're loading a CuDNNLSTM layer
      # from the number of bias weights:
      # CuDNNLSTM has (units * 8) weights; while LSTM has (units * 4)
      # if there's no bias weight in the file, skip this conversion
      units = weights[1].shape[0]
      bias_shape = weights[2].shape
      n_gates = 4

      if bias_shape == (2 * units * n_gates,):
        source = 'CuDNNLSTM'
      elif bias_shape == (units * n_gates,):
        source = 'LSTM'
      else:
        raise ValueError('Invalid bias shape: ' + str(bias_shape))

      def convert_weights(weights, from_cudnn=True):
        # transpose (and reshape) input and recurrent kernels
        kernels = transform_kernels(weights[0],
                                    transpose_input(from_cudnn),
                                    n_gates)
        recurrent_kernels = transform_kernels(weights[1], lambda k: k.T, n_gates)
        if from_cudnn:
          # merge input and recurrent biases into a single set
          biases = np.sum(np.split(weights[2], 2, axis=0), axis=0)
        else:
          # Split single set of biases evenly to two sets. The way of
          # splitting doesn't matter as long as the two sets sum is kept.
          biases = np.tile(0.5 * weights[2], 2)
        return [kernels, recurrent_kernels, biases]

      if source != target_class:
        weights = convert_weights(weights, from_cudnn=source == 'CuDNNLSTM')
    # end if LSTM

    # convert the weights between CuDNNGRU and GRU(reset_after=True)
    if target_class in ['GRU', 'CuDNNGRU'] and len(weights) == 3:
      # We can determine the source of the weights from the shape of the bias.
      # If there is no bias we skip the conversion
      # since CuDNNGRU always has biases.

      units = weights[1].shape[0]
      bias_shape = weights[2].shape
      n_gates = 3

      def convert_weights(weights, from_cudnn=True):
        kernels = transform_kernels(weights[0],
                                    transpose_input(from_cudnn),
                                    n_gates)
        recurrent_kernels = transform_kernels(weights[1], lambda k: k.T, n_gates)
        biases = np.array(weights[2]).reshape((2, -1) if from_cudnn else -1)
        return [kernels, recurrent_kernels, biases]

      if bias_shape == (2 * units * n_gates,):
        source = 'CuDNNGRU'
      elif bias_shape == (2, units * n_gates):
        source = 'GRU(reset_after=True)'
      elif bias_shape == (units * n_gates,):
        source = 'GRU(reset_after=False)'
      else:
        raise ValueError('Invalid bias shape: ' + str(bias_shape))

      if target_class == 'CuDNNGRU':
        target = 'CuDNNGRU'
      elif layer.reset_after:
        target = 'GRU(reset_after=True)'
      else:
        target = 'GRU(reset_after=False)'

      # only convert between different types
      if source != target:
        # types = (source, target)
        # if 'GRU(reset_after=False)' in types:
        #    raise ValueError('%s is not compatible with %s' % types)
        if source == 'CuDNNGRU':
          weights = convert_weights(weights, from_cudnn=True)
        elif source == 'GRU(reset_after=True)':
          weights = convert_weights(weights, from_cudnn=False)
    # end if GRU

    return weights

  @staticmethod
  def evaluate_summary(model, res):
    """
    inputs:
      model: tf.keras model
      res: results from `evaluate` or `train_on_batch` calls

    returns:
      string with the summary info
    """
    if type(res) not in [tuple, list, np.ndarray]:
      res = [res]
    metrics = model.metrics_names
    s_res = ""
    for i, metric in enumerate(metrics):
      s_res += "{}: {:.4f}  ".format(metric, res[i])
    s_res = s_res[:-1]
    return s_res

  def tfdataset_to_generator(self, tf_ds):
    """
    this function will take a prepared tf.data.Dataset
    and will create a infinite generator out of it
    Observations:
      The `tf_ds` prepared Dataset should have a predefined .batch()
      and also preferably a .prefetch()
    """
    import tensorflow as tf
    if tf.executing_eagerly():
      while True:
        for batch in tf_ds:
          yield batch
    else:
      tf_iter = tf.compat.v1.data.make_initializable_iterator(tf_ds)
      tf_batch = tf_iter.get_next()
      tf_data_init = tf_iter.initializer
      sess = tf.compat.v1.keras.backend.get_session()
      sess.run(tf_data_init)
      gen_iter = 0
      while True:
        try:
          batch = sess.run(tf_batch)
          gen_iter += 1
          yield batch
        except:
          sess.run(tf_data_init)
          self.P("Dataset finished as iter {}. Resetting iterator".format(gen_iter))
          gen_iter = 0

  def check_dataset_datetime_tensors(self, x_tensors, idx_md, idx_m, idx_wd, year,
                                     offset_month=1, offset_day=1):
    md = int(x_tensors[idx_md][0, 0, 0]) + offset_day
    wd = int(x_tensors[idx_wd][0, 0, 0])
    m = int(x_tensors[idx_m][0, 0, 0]) + offset_month
    first_day = dt(year=year, month=m, day=md)
    if wd != first_day.weekday():
      self.raise_error("Week day {} of start date {} differs from dataset {}".format(
        first_day.weekday(), first_day, wd))
    else:
      self.P("Dataset datetime start check ok: start Y:{} M:{} D:{} (wd:{}) on day {}".format(
        year, m, md, first_day.weekday(), wd))
    return

  @staticmethod
  def get_tf_optimizer(str_optimizer):
    import tensorflow.compat.v1 as tf
    str_optimizers = ['rmsprop', 'sgd', 'adam', 'nadam', 'adagrad']
    assert str_optimizer in str_optimizers

    if str_optimizer == 'rmsprop':
      return tf.train.RMSPropOptimizer
    if str_optimizer == 'sgd':
      return tf.train.GradientDescentOptimizer
    if str_optimizer == 'adam':
      return tf.train.AdamOptimizer
    if str_optimizer == 'nadam':
      return tf.contrib.opt.NadamOptimizer
    if str_optimizer == 'adagrad':
      return tf.train.AdagradOptimizer

  def save_optimizer_weights(self, optimizer, fn, subfolder_path=None):
    import tensorflow as tf
    symbolic_weights = getattr(optimizer, 'weights')
    weight_values = tf.keras.backend.batch_get_value(symbolic_weights)
    return self.save_pickle_to_models(data=weight_values, fn=fn, subfolder_path=subfolder_path)

  @staticmethod
  def safe_keras_predict(model, X, batch_size=None, verbose=0):
    import tensorflow as tf

    if batch_size is None:
      batch_size = 32

    if not isinstance(X, list):
      X = [X]

    is_numpy = isinstance(X[0], np.ndarray)

    nr_observations_per_tensor = [x.shape[0] for x in X]
    assert len(set(nr_observations_per_tensor)) == 1
    nr_obeservations = nr_observations_per_tensor[0]

    nr_batches = nr_obeservations // batch_size
    if nr_obeservations % batch_size != 0:
      nr_batches += 1

    R = range(nr_batches)
    if verbose > 0:
      from tqdm import tqdm
      R = tqdm(R)

    y_hat = None
    for batch_step in R:
      start = batch_step * batch_size
      end = (batch_step + 1) * batch_size
      X_batch = [x[start:end] for x in X]

      if is_numpy:
        X_batch = [tf.convert_to_tensor(x) for x in X_batch]

      crt_y_hat = model(X_batch)

      if not isinstance(crt_y_hat, list):
        crt_y_hat = [crt_y_hat]

      nr_outputs = len(crt_y_hat)
      crt_y_hat = [y.numpy() for y in crt_y_hat]

      if y_hat is None:
        y_hat = [[] for _ in range(nr_outputs)]

      for idx_output in range(nr_outputs):
        y_hat[idx_output].append(crt_y_hat[idx_output])
    # endfor

    y_hat = [np.concatenate(y) for y in y_hat]
    if len(y_hat) == 1:
      y_hat = y_hat[0]

    return y_hat

  def compare_keras_models(self, model1, model2):
    layers1 = model1.layers
    layers2 = model2.layers
    _equal = True
    if len(layers1) != len(layers2):
      self.P("Number of layers differs!")
      _equal = False
    if _equal:
      for i in range(len(layers1)):
        w1_list = layers1[i].get_weights()
        w2_list = layers2[i].get_weights()
        for j in range(len(w1_list)):
          w1 = w1_list[j]
          w2 = w2_list[j]
          diff = np.sum((w1 == w2) == False)
          if diff > 0:
            self.P("Found {} diffs in layers {}/{} on weights:{}".format(
              diff, layers1[i].name, layers2[i].name, j
            ))
            _equal = False
    if not _equal:
      self.P("Models comparison failed!")
    else:
      self.P("Models are equal...")
    return _equal


import tensorflow as tf

def keras_model_memory_usage_in_bytes(model, *, batch_size: int):
    """
    Return the estimated memory usage of a given Keras model in bytes.
    This includes the model weights and layers, but excludes the dataset.

    The model shapes are multipled by the batch size, but the weights are not.

    Args:
        model: A Keras model.
        batch_size: The batch size you intend to run the model with. If you
            have already specified the batch size in the model itself, then
            pass `1` as the argument here.
    Returns:
        An estimate of the Keras model's memory usage in bytes.

    """
    default_dtype = tf.keras.backend.floatx()
    shapes_mem_count = 0
    internal_model_mem_count = 0
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            internal_model_mem_count += keras_model_memory_usage_in_bytes(
                layer, batch_size=batch_size
            )
        single_layer_mem = tf.as_dtype(layer.dtype or default_dtype).size
        out_shape = layer.output_shape
        if isinstance(out_shape, list):
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem

    trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.trainable_weights]
    )
    non_trainable_count = sum(
        [tf.keras.backend.count_params(p) for p in model.non_trainable_weights]
    )

    total_memory = (
        batch_size * shapes_mem_count
        + internal_model_mem_count
        + trainable_count
        + non_trainable_count
    )
    return total_memory


class JsonUtil:
  @staticmethod
  def compare_json(left, right, keys_to_ignore=None) -> bool:
    """
    Compare two json strings
    Parameters
    ----------
    left
    right
    keys_to_ignore

    Returns
    -------
    bool
    """
    if keys_to_ignore is None:
      keys_to_ignore = []
    if left is None and right is None:
      return True
    elif left is None or right is None:
      return False
    elif isinstance(left, str) and isinstance(right, str):
      return left == right
    elif isinstance(left, list) and isinstance(right, list):
      return JsonUtil.__compare_json_list(left, right, keys_to_ignore)
    elif isinstance(left, dict) and isinstance(right, dict):
      return JsonUtil.__compare_json_dict(left, right, keys_to_ignore)
    else:
      return False

  @staticmethod
  def __compare_json_list(left, right, keys_to_ignore):
    """
    Compare two json lists
    Parameters
    ----------
    left
    right
    keys_to_ignore

    Returns
    -------
    bool
    """
    if len(left) != len(right):
      return False
    for i in range(len(left)):
      if not JsonUtil.compare_json(left[i], right[i], keys_to_ignore):
        return False
    return True

  @staticmethod
  def __compare_json_dict(left, right, keys_to_ignore):
    """
    Compare two json dicts
    Parameters
    ----------
    left
    right
    keys_to_ignore

    Returns
    -------
    bool
    """
    if len(left) != len(right):
      return False
    for key in left:
      if key not in right:
          return False
      if key in keys_to_ignore:
          continue
      if not left[key] == right[key]:
        return False
    return True



import os
import numpy as np

### TODO flatten2dlist

class _NLPMixin(object):
  """
  Mixin for NLP functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_HistogramMixin`:
    - self.show_text_histogram
  """

  def __init__(self):
    super(_NLPMixin, self).__init__()

    try:
      from .histogram_mixin import _HistogramMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _NLPMixin without having _HistogramMixin")

    return

  def _text_to_observation(self, sent, tokenizer_func, max_size, dct_word2idx,
                           unk_word_func=None,
                           get_embeddings=True, embeddings=None,
                           PAD_ID=0, UNK_ID=1, left_pad=False, cut_left=False):
    if type(sent) != str:
      self.raise_error("sent must be a string")
    if unk_word_func is None:
      unk_word_func = lambda x: UNK_ID
    ids, tokens = tokenizer_func(sent, dct_word2idx, unk_word_func)
    ids = ids[-max_size:] if cut_left else ids[:max_size]
    n_unks = ids.count(UNK_ID)

    all_unk_indices = [i for i, x in enumerate(ids) if x == UNK_ID]
    all_unk_words = set([tokens[i] for i in all_unk_indices])

    if len(ids) < max_size:
      if left_pad:
        ids = [PAD_ID] * (max_size - len(ids)) + ids
      else:
        ids = ids + [PAD_ID] * (max_size - len(ids))
    if get_embeddings:
      np_output = np.array([embeddings[i] for i in ids])
    else:
      np_output = np.array(ids)
    return np_output, n_unks, list(all_unk_words)

  def corpus_to_batch(self, sents, tokenizer_func, max_size, dct_word2idx,
                      get_embeddings=True, embeddings=None,
                      unk_word_func=None,
                      PAD_ID=0, UNK_ID=1, left_pad=False, cut_left=False):
    """
    sents: list of sentence
    dct_word2idx : word to word-id dict
    tokenizer_func: function with signature (sentence, dict, unk_word_tokenizer_func)
    max_size : max obs size
    embeddinds :  matrix
    get_embeddings: return embeds no ids
    PAD_ID : pad id
    UNK_ID : unknown word id
    left_pad : pad to the left
    cur_left : cut to the left
    """
    if type(sents) != list or type(sents[0]) != str:
      self.raise_error("sents must be a list of strings")

    result = [
      self._text_to_observation(
        sent=x,
        tokenizer_func=tokenizer_func,
        max_size=max_size,
        dct_word2idx=dct_word2idx,
        unk_word_func=unk_word_func,
        get_embeddings=get_embeddings,
        embeddings=embeddings,
        PAD_ID=PAD_ID,
        UNK_ID=UNK_ID,
        left_pad=left_pad,
        cut_left=cut_left
      )
      for x in sents
    ]

    output, n_unks, all_unk_words = list(zip(*result))
    np_batch = np.array(list(output))
    self.P("Tokenized {} sentences to {}. Found {} unknown words: {}".format(
      len(sents), np_batch.shape, sum(n_unks), self.flatten_2d_list(list(all_unk_words))))
    return np_batch

  def load_documents(self, folder, doc_ext='.txt', label_ext='.label',
                     doc_folder=None, label_folder=None,
                     return_labels_list=True, exclude_list=[],
                     min_label_freq=0):
    """
     loads either from a _data folder or from other folder all documents and
     their labels for NLP tasks
     doc_ext : the documents extension
     label_ext : the label files extension
     doc_folder, label_folder : if documents and labels are separated...

     returns:
       list of documents (strings)
       list of labels (lists)
    """

    path = self.get_data_subfolder(folder)
    if path is None:
      path = folder if os.path.isdir(folder) else None
    if path is None:
      raise ValueError("Cannot find folder '{}'".format(path))

    if doc_folder is not None:
      path_docs = os.path.join(path, doc_folder)
    else:
      path_docs = path

    if label_folder is not None:
      path_labels = os.path.join(path, label_folder)
    else:
      path_labels = path

    if not os.path.isdir(path_docs):
      raise ValueError("Cannot find folder '{}'".format(path_docs))
    if not os.path.isdir(path_labels):
      raise ValueError("Cannot find folder '{}'".format(path_labels))

    self.P("Searching docs in '...{}' ...".format(path_docs[-25:]))
    all_files = os.listdir(path_docs)
    n_labels = 0
    lst_docs = []
    lst_labels = []
    dct_labels = {}
    set_labels = set()

    for fn in all_files:
      if fn[-len(doc_ext):] == doc_ext:
        fn_full = os.path.join(path_docs, fn)
        fn_name = fn[:-len(doc_ext)]
        fn_label = fn_name + label_ext
        fn_label_full = os.path.join(path_labels, fn_label)
        txt_doc = None
        txt_labels = None
        with open(fn_full, 'rt', encoding='utf-8') as f_doc:
          txt_doc = f_doc.read()
        if os.path.isfile(fn_label_full):
          with open(fn_label_full, 'rt') as f_lab:
            txt_labels = f_lab.readlines()
            txt_labels = [x.replace('\n', '') for x in txt_labels]
            for _str in exclude_list:
              txt_labels = [x.replace(_str, '') for x in txt_labels]
            txt_labels = list(filter(lambda x: x != '', txt_labels))
            for txt_label in txt_labels:
              if txt_label in dct_labels.keys():
                dct_labels[txt_label] += 1
              else:
                dct_labels[txt_label] = 1
            n_labels += len(txt_labels)
            set_labels.update(txt_labels)
        lst_docs.append(txt_doc)
        lst_labels.append(txt_labels)
    n_lab_files = len(lst_labels) - lst_labels.count(None)

    occurences = list(dct_labels.values())
    min_occ = np.min(occurences)
    max_occ = np.max(occurences)
    avg_occ = np.mean(occurences)
    med_occ = np.median(occurences)

    self.P("Loaded {} docs, {} lbl files:".format(len(lst_docs), n_lab_files))
    self.P("  {} unique labels with occurence/doc min/max/avg/med: {:.1f}/{:.1f}/{:.1f}/{:.1f}".format(
      len(set_labels), min_occ, max_occ, avg_occ, med_occ))
    self.P("  Overall {:.1f} lbl/doc and {} unlabeled docs".format(
      n_labels / len(lst_docs),
      lst_labels.count(None)))

    if min_label_freq > 0:
      dct_new_labels = {k: v for k, v in dct_labels.items() if v >= min_label_freq}
      set_new_labels = set_labels & set(list(dct_new_labels.keys()))
      lst_new_labels = []
      n_labels = 0
      for l in lst_labels:
        lst_new_labels.append(list(filter(lambda x: x in dct_new_labels.keys(), l)))
        n_labels += len(lst_new_labels[-1])

      lst_labels = lst_new_labels
      set_labels = set_new_labels

      occurences = list(dct_new_labels.values())
      min_occ = np.min(occurences)
      max_occ = np.max(occurences)
      avg_occ = np.mean(occurences)
      med_occ = np.median(occurences)

      self.P("  [after processing min label freq = {}]:".format(min_label_freq))
      self.P("  {} unique labels with occurence/doc min/max/avg/med: {:.1f}/{:.1f}/{:.1f}/{:.1f}".format(
        len(set_labels), min_occ, max_occ, avg_occ, med_occ))
      self.P("  Overall {:.1f} lbl/doc and {} unlabeled docs".format(
        n_labels / len(lst_docs),
        lst_labels.count(None)))
    # endif

    self.show_text_histogram(
      data=[len(x) for x in lst_docs],
      caption='Documents length (chars) distribution',
      show_both_ends=True
    )

    if return_labels_list:
      return lst_docs, lst_labels, list(set_labels)
    else:
      return lst_docs, lst_labels



import cv2
import os
import traceback
from naeural_core import constants as ct
from time import time


def get_rotate_code_from_orientation(cap_orientation):
  rotate_code = None
  if int(cap_orientation) == 270:
    rotate_code = cv2.ROTATE_90_CLOCKWISE
  elif int(cap_orientation) == 180:
    rotate_code = cv2.ROTATE_180
  elif int(cap_orientation) == 90:
    rotate_code = cv2.ROTATE_90_COUNTERCLOCKWISE

  return rotate_code


class _VideoFileMixin(object):

  def __init__(self):
    self._path = None
    self._progress_interval = list(range(0, 101, 10))

    self._cv2_cap = None
    self._cv2_orientation = None
    self._cv2_rotate_code = None
    self._decord_vr = None

    super(_VideoFileMixin, self).__init__()
    return

  @property
  def cfg_notify_download_downstream(self):
    return self.cfg_stream_config_metadata.get('NOTIFY_DOWNLOAD_DOWNSTREAM', False)

  def _maybe_download(self):
    if os.path.exists(str(self.video_file_url)):
      self.P('Found local file.')
      self._path = self.video_file_url

      self._metadata.download = 100
      self._metadata.download_start_date = time()
      self._metadata.download_elapsed_time = time() - self._metadata.download_start_date
      if self.cfg_notify_download_downstream:
        metadata = self._metadata.__dict__.copy()
        self._add_inputs(
          [
            self._new_input(struct_data=metadata, metadata=metadata),
          ]
        )
      self.P('File to be used: {}'.format(self._path))
    else:
      self.P('Download...')
      try:
        self._path = self._download(
          url=self.video_file_url,
          progress=False,
          verbose=False,
          notify_download=self._notify_download_progress,
        )
        self.P('File downloaded in: {}'.format(self._path))
      except:
        msg = "Exception download"
        info = traceback.format_exc()
        self.P(msg + "\n" + info, color='y')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          msg=msg,
          info=info
        )
      #end try-except
    #endif

    self._path = os.path.abspath(self._path)
    if hasattr(self._metadata, 'path'):
      self._metadata.path = self._path

    return

  def _notify_download_progress(self, message):
    try:
      prc = message.split(':')[1]
      prc = prc.replace('%', '')
      prc = int(float(prc))
      if prc % 10 == 0 and prc in self._progress_interval:
        if prc == 0:
          vars(self._metadata)['download_start_date'] = time()

        vars(self._metadata)['download'] = prc
        vars(self._metadata)['download_elapsed_time'] = time() - self._metadata.download_start_date
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_NORMAL,
          msg='Download status',
          info='{}'.format(self._metadata.__dict__),
          displayed=True,
        )
        if int(prc) in [10, 30, 70, 100]:
          self.P("Download status: {}%".format(prc))
        self._progress_interval.remove(prc)

        if self.cfg_notify_download_downstream:
          metadata = self._metadata.__dict__.copy()
          self._add_inputs(
            [
              self._new_input(struct_data=metadata, metadata=metadata),
            ]
          )
        # endif
    except Exception as e:
      self.P('Exception while notify progress: {}'.format(str(e)), color='r')
    return

  def _create_cv2_capture(self, path=None):
    path = path or self._path
    
    if self.cfg_use_ffmpeg:
      self.P("Opening capture with CAP_FFMPEG...")
      self._cv2_cap = cv2.VideoCapture(path, cv2.CAP_FFMPEG)
    else:
      self._cv2_cap = cv2.VideoCapture(path)
      
    self.cv2_supports_rotation = True
    try:
      if self.cfg_custom_rotation:
        self._cv2_cap.set(cv2.CAP_PROP_ORIENTATION_AUTO, 0)
      self._cv2_orientation = self._cv2_cap.get(cv2.CAP_PROP_ORIENTATION_META)
      self._cv2_rotate_code = get_rotate_code_from_orientation(self._cv2_orientation)
    except Exception as e:
      self.cv2_supports_rotation = False
      msg = "Video File plugin cv2 does not support orientation API"
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info=str(e),
        displayed=True,
      )
      self.P(msg, color='r')
      
    self._h, self._w = int(self._cv2_cap.get(cv2.CAP_PROP_FRAME_HEIGHT)), int(self._cv2_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    if self.cv2_supports_rotation and self._cv2_rotate_code in [0, 2]:
      self._h, self._w = self._w, self._h

    return

  def _create_decord_video_reader(self, path=None):
    path = path or self._path
    
    try:
      from decord import VideoReader, cpu
    except ModuleNotFoundError:
      msg = "ERROR: Missing decord package renders unusable current DCT!"
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        displayed=True,
      )
      return

    try:
      self._decord_vr = VideoReader(path, ctx=cpu(0))
    except:
      pass

    return

  def _release_cv2_capture(self):
    if self._cv2_cap is None:
      return

    self._cv2_cap.release()
    del self._cv2_cap
    self._cv2_cap = None
    return

  def _release_decord_video_reader(self):
    if self._decord_vr is None:
      return

    del self._decord_vr
    self._decord_vr = None
    return

  def _reset_cv2_capture(self, path=None):
    self._release_cv2_capture()
    self._create_cv2_capture(path=path)
    return

  def _reset_decord_video_reader(self, path=None):
    self._release_decord_video_reader()
    self._create_decord_video_reader(path=path)
    assert self._decord_vr is not None, "`_create_decord_video_reader` failed - cannot use current DCT"
    return

  def _get_number_of_frames(self):
    # frame_count = int(self._capture.get(cv2.CAP_PROP_FRAME_COUNT))
    # cv2 doesn't always return the true number of frames. in order to
    # be sure that we report the correct number of frames, we will count the number of frames by reading the entire movie

    if self._decord_vr is not None:
      return len(self._decord_vr)
    # TODO:
    #  Find a efficient solution for this !!!!
    i = 0
    while True:
      has_frame = self._cv2_cap.grab()
      if not has_frame:
        break
      i += 1
    # self._capture.set(cv2.CAP_PROP_POS_FRAMES, 0)
    return i





from .commandcontrol_comm_mixin import _CommandControlCommMixin
from .default_comm_mixin import _DefaultCommMixin
from .heartbeats_comm_mixin import _HeartbeatsCommMixin
from .notifications_comm_mixin import _NotificationsCommMixin
from .telemetry_mixin import _CommunicationTelemetryMixin



import cv2
import numpy as np
import matplotlib.pyplot as plt

# targeted rectangle on original image which needs to be transformed
#vlcsnap-2021-07-02-13h26m43s659.png
tl = [533, 310]
tr = [674, 267]
bl = [792, 464]
br = [971, 393]

#vlcsnap-2021-07-02-15h29m06s352.png
# tl = [822, 220]
# tr = [928, 193]
# bl = [1144, 322]
# br = [1244, 282]

#'/home/work/Pictures/vlcsnap-2021-07-02-12h51m02s764.png'
# tl = [1017, 359]
# tr = [1112, 333]
# bl = [1204, 421]
# br = [1291, 391]


OX_center = int((tl[0] + tr[0] + bl[0] + br[0]) / 4)
OY_center = int((tl[1] + tr[1] + bl[1] + br[1]) / 4)

corner_points_array = np.float32([tl,tr,br,bl])

# original image dimensions
width = 1920
height = 1080
img = cv2.imread('/home/work/Pictures/vlcsnap-2021-07-02-15h29m06s352.png')
# img = cv2.imread('/home/work/Pictures/vheat.png')
# img = cv2.imread('/home/work/Pictures/vlcsnap-2021-07-02-15h29m06s352.png')
# print(np.shape(img))
img = cv2.resize(img,(1920,1080))

# Create an array with the parameters (the dimensions) required to build the matrix


width_AD = np.sqrt(((tl[0] - tr[0]) ** 2) + ((tl[1] - tr[1]) ** 2))
width_BC = np.sqrt(((bl[0] - br[0]) ** 2) + ((bl[1] - br[1]) ** 2))
maxWidth = max(int(width_AD), int(width_BC))


height_AB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
height_CD = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
maxHeight = max(int(height_AB), int(height_CD))



# imgTl = [OX_center + 0, OY_center + 0]
# imgTr = [OX_center + width, OY_center + 0]
# imgBr = [OX_center + width, OY_center + height]
# imgBl = [OX_center + 0, OY_center + height]

imgTl = [0, 0]
imgTr = [ width, 0]
imgBr = [width,  height]
imgBl = [ 0,  height]
img_params = np.float32([imgTl,imgTr,imgBr,imgBl])

# input_pts = np.float32([pt_A, pt_B, pt_C, pt_D])
output_pts = np.float32([[OX_center + 0, OY_center + 0],
                        [OX_center + maxWidth, OY_center + 0],
                        [OX_center + maxWidth, OY_center + maxHeight],
                        [OX_center + 0, OY_center + maxHeight]])

# Compute and return the transformation matrix
matrix = cv2.getPerspectiveTransform(corner_points_array,output_pts)

img_transformed = np.zeros((height, width, 3), np.uint8)
print(np.shape(img))
print(np.shape(img_transformed))
img_transformed[:] = (0, 0, 0)

# img_transformed = cv2.warpPerspective(img,matrix,(width, height), borderMode=cv2.BORDER_CONSTANT, borderValue=(255, 0, 0))
# print(img_transformed)
count = 0
for i in range(width):
  for j in range(height):
    coord = [i, j, 1]
    new_coord = matrix@coord
    if int(new_coord[0] / new_coord[2]) >= 0 and int(new_coord[0] / new_coord[2]) < width and int(new_coord[1] / new_coord[2]) >= 0 and int(new_coord[1] / new_coord[2])< height:
      count += 1
      img_transformed[int(new_coord[1] / new_coord[2]), int(new_coord[0] / new_coord[2])] = img[j,i]





print(count)


plt.imshow(img_transformed[:, :, ::-1])  # Show results
# cv2.imwrite('/home/work/Pictures/BUC_top.png', img_transformed)
plt.show()